from typing import AsyncGenerator

from pymobiledevice3.remote.core_device.core_device_service import CoreDeviceService
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService


class DiagnosticsServiceService(CoreDeviceService):
    """
    Obtain device diagnostics
    """

    SERVICE_NAME = 'com.apple.coredevice.diagnosticsservice'

    def __init__(self, rsd: RemoteServiceDiscoveryService):
        super().__init__(rsd, self.SERVICE_NAME)

    async def capture_sysdiagnose(self, is_dry_run: bool) -> AsyncGenerator[bytes, None]:
        response = await self.invoke('com.apple.coredevice.feature.capturesysdiagnose', {'isDryRun': is_dry_run})
        return self.service.iter_file_chunks(response['fileTransfer']['expectedLength'])



import logging
from functools import cached_property
from typing import Any, Mapping, Optional

from pymobiledevice3 import usbmux
from pymobiledevice3.exceptions import ConnectionFailedError, NoDeviceConnectedError
from pymobiledevice3.restore.restore_options import RestoreOptions
from pymobiledevice3.service_connection import ServiceConnection


class RestoredClient:
    DEFAULT_CLIENT_NAME = 'pyMobileDevice'
    SERVICE_PORT = 62078

    def __init__(self, udid: Optional[str] = None) -> None:
        self.logger = logging.getLogger(__name__)
        self.udid = udid
        self.version: Optional[str] = None
        self.query_type: Optional[str] = None
        self.label: Optional[str] = None
        self.service: Optional[ServiceConnection] = None

    async def connect(self, client_name: str = DEFAULT_CLIENT_NAME) -> None:
        self.service = ServiceConnection.create_using_usbmux(self.udid, self.SERVICE_PORT,
                                                             connection_type='USB')
        await self.service.aio_start()
        self.label = client_name
        self.query_type = await self.service.aio_send_recv_plist({'Request': 'QueryType'})
        self.version = self.query_type.get('RestoreProtocolVersion')

        assert self.query_type.get('Type') == 'com.apple.mobile.restored', f'wrong query type: {self.query_type}'

    @staticmethod
    def _get_or_verify_udid(udid: Optional[str] = None) -> str:
        device = usbmux.select_device(udid)
        if device is None:
            if udid:
                raise ConnectionFailedError()
            else:
                raise NoDeviceConnectedError()
        return device.serial

    async def query_value(self, key: Optional[str] = None) -> Any:
        req = {'Request': 'QueryValue', 'Label': self.label}

        if key:
            req['QueryKey'] = key

        return await self.service.aio_send_recv_plist(req)

    async def start_restore(self, opts: Optional[RestoreOptions] = None) -> None:
        req = {'Request': 'StartRestore', 'Label': self.label, 'RestoreProtocolVersion': self.version}

        if opts is not None:
            req['RestoreOptions'] = opts.to_dict()

        self.logger.debug(f'start_restore request: {req}')

        return await self.service.aio_send_plist(req)

    async def reboot(self) -> Mapping:
        return await self.service.aio_send_recv_plist({'Request': 'Reboot', 'Label': self.label})

    async def send(self, message: Mapping) -> None:
        await self.service.aio_send_plist(message)

    async def recv(self) -> Mapping:
        return await self.service.aio_recv_plist()

    @cached_property
    async def hardware_info(self) -> Mapping[str, Any]:
        return (await self.query_value('HardwareInfo'))['HardwareInfo']

    @property
    async def saved_debug_info(self) -> Mapping[str, Any]:
        return (await self.query_value('SavedDebugInfo'))['SavedDebugInfo']



import logging
import random
import time
from abc import abstractmethod

import gpxpy


class LocationSimulationBase:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    @abstractmethod
    def set(self, latitude: float, longitude: float) -> None:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass

    def play_gpx_file(self, filename: str, disable_sleep: bool = False, timing_randomness_range: int = 0):
        with open(filename) as f:
            gpx = gpxpy.parse(f)

        last_time = None
        gpx_timing_noise = None
        for track in gpx.tracks:
            for segment in track.segments:
                for point in segment.points:
                    if last_time is not None:
                        duration = (point.time - last_time).total_seconds()
                        if duration >= 0:
                            if not disable_sleep:

                                if timing_randomness_range:
                                    gpx_timing_noise = random.randint(-timing_randomness_range, timing_randomness_range) / 1000
                                    duration += gpx_timing_noise

                                self.logger.info(f'waiting for {duration:.3f}s')
                                time.sleep(duration)
                    last_time = point.time
                    self.logger.info(f'set location to {point.latitude} {point.longitude}')
                    self.set(point.latitude, point.longitude)



import time

import click

from pymobiledevice3.cli.cli_common import Command
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.power_assertion import PowerAssertionService


@click.group()
def cli() -> None:
    pass


@cli.command('power-assertion', cls=Command)
@click.argument('type', type=click.Choice(
    ['AMDPowerAssertionTypeWirelessSync', 'PreventUserIdleSystemSleep', 'PreventSystemSleep']))
@click.argument('name')
@click.argument('timeout', type=click.INT)
@click.argument('details', required=False)
def power_assertion(service_provider: LockdownServiceProvider, type, name, timeout, details) -> None:
    """ Create a power assertion """
    with PowerAssertionService(service_provider).create_power_assertion(type, name, timeout, details):
        print('> Hit Ctrl+C to exit')
        time.sleep(timeout)



from packaging.version import Version

from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.services.remote_server import RemoteServer


class DvtTestmanagedProxyService(RemoteServer):
    SERVICE_NAME = 'com.apple.testmanagerd.lockdown.secure'
    OLD_SERVICE_NAME = 'com.apple.testmanagerd.lockdown'
    RSD_SERVICE_NAME = 'com.apple.dt.testmanagerd.remote'

    # TODO: there is also service named 'com.apple.dt.testmanagerd.remote.automation', but not used

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, RemoteServiceDiscoveryService):  # only happends when >=17.0
            service_name = self.RSD_SERVICE_NAME
            remove_ssl_context = False
        elif Version(lockdown.product_version) >= Version('14.0'):
            service_name = self.SERVICE_NAME
            remove_ssl_context = False
        else:
            service_name = self.OLD_SERVICE_NAME
            remove_ssl_context = True

        super().__init__(lockdown, service_name, remove_ssl_context=remove_ssl_context)



import socket
from typing import Generator, Mapping, Union

from pymobiledevice3.exceptions import NotificationTimeoutError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.services.lockdown_service import LockdownService


class NotificationProxyService(LockdownService):
    SERVICE_NAME = 'com.apple.mobile.notification_proxy'
    RSD_SERVICE_NAME = 'com.apple.mobile.notification_proxy.shim.remote'

    INSECURE_SERVICE_NAME = 'com.apple.mobile.insecure_notification_proxy'
    RSD_INSECURE_SERVICE_NAME = 'com.apple.mobile.insecure_notification_proxy.shim.remote'

    def __init__(self, lockdown: LockdownServiceProvider, insecure=False, timeout: Union[float, int] = None):
        if isinstance(lockdown, RemoteServiceDiscoveryService):
            secure_service_name = self.RSD_SERVICE_NAME
            insecure_service_name = self.RSD_INSECURE_SERVICE_NAME
        else:
            secure_service_name = self.SERVICE_NAME
            insecure_service_name = self.INSECURE_SERVICE_NAME

        if insecure:
            super().__init__(lockdown, insecure_service_name)
        else:
            super().__init__(lockdown, secure_service_name)

        if timeout is not None:
            self.service.socket.settimeout(timeout)

    def notify_post(self, name: str) -> None:
        """ Send notification to the device's notification_proxy. """
        self.service.send_plist({'Command': 'PostNotification', 'Name': name})

    def notify_register_dispatch(self, name: str) -> None:
        """ Tells the device to send a notification on the specified event. """
        self.logger.info(f'Observing {name}')
        self.service.send_plist({'Command': 'ObserveNotification', 'Name': name})

    def receive_notification(self) -> Generator[Mapping, None, None]:
        while True:
            try:
                yield self.service.recv_plist()
            except socket.timeout as e:
                raise NotificationTimeoutError from e



import time

LOCKDOWND_SOCKET_SELECT_TIMEOUT = 60


def test_lockdown_reconnect(lockdown):
    d1 = lockdown.date

    # add some threshold to make sure lockdownd closed the connection on its end
    time.sleep(LOCKDOWND_SOCKET_SELECT_TIMEOUT + 5)

    d2 = lockdown.date

    assert d1 < d2



import time
from pathlib import Path
from ssl import SSLEOFError
from typing import Callable

import pytest

from pymobiledevice3.exceptions import ConnectionFailedError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.mobilebackup2 import Mobilebackup2Service

PASSWORD = '1234'


def ignore_connection_errors(f: Callable):
    """
    The device may become unresponsive for a short while after changing the password settings and reject
    incoming connections at different stages
    """

    def _wrapper(*args, **kwargs):
        while True:
            try:
                f(*args, **kwargs)
                break
            except (SSLEOFError, ConnectionAbortedError, OSError, ConnectionFailedError):
                time.sleep(1)

    return _wrapper


@ignore_connection_errors
def change_password(lockdown, old: str = '', new: str = '') -> None:
    with Mobilebackup2Service(lockdown) as service:
        service.change_password(old=old, new=new)


@ignore_connection_errors
def backup(lockdown: LockdownClient, backup_directory: Path) -> None:
    with Mobilebackup2Service(lockdown) as service:
        service.backup(full=True, backup_directory=backup_directory)


@pytest.mark.filterwarnings('ignore::UserWarning')
def test_backup(lockdown, tmp_path):
    backup(lockdown, tmp_path)


@pytest.mark.filterwarnings('ignore::UserWarning')
def test_encrypted_backup(lockdown, tmp_path):
    change_password(lockdown, new=PASSWORD)
    backup(lockdown, tmp_path)
    change_password(lockdown, old=PASSWORD)



import pathlib
from datetime import datetime

import pytest

from pymobiledevice3.exceptions import AfcException, AfcFileNotFoundError
from pymobiledevice3.services.afc import MAXIMUM_READ_SIZE, AfcService, afc_error_t

TEST_FILENAME = 'test'
TEST_FOLDER_NAME = 'test_folder'


@pytest.fixture(scope='function')
def afc(lockdown):
    with AfcService(lockdown) as afc:
        yield afc


def test_exists(afc: AfcService):
    assert afc.exists('DCIM')
    assert not afc.exists('a_directory_that_doesnt_exist')


def test_exists_folder_in_a_file(afc: AfcService):
    afc.set_file_contents(TEST_FILENAME, b'data')
    try:
        assert not afc.exists(f'{TEST_FILENAME}/sub_folder')
    finally:
        afc.rm(TEST_FILENAME)


def test_rm(afc):
    afc.set_file_contents(TEST_FILENAME, b'')
    filenames = afc.listdir('/')
    assert TEST_FILENAME in filenames

    afc.rm(TEST_FILENAME)

    filenames = afc.listdir('/')
    assert TEST_FILENAME not in filenames


def test_rm_force_missing_file(afc):
    with pytest.raises(AfcFileNotFoundError):
        afc.rm(TEST_FILENAME)
    afc.rm(TEST_FILENAME, force=True)


@pytest.mark.parametrize('path', [
    'file_that_doesnt_exist.txt',
    'missingfolder/file_that_doesnt_exist.txt',
    '/missingfolder/file_that_doesnt_exist.txt',
    '/missingfolder/file_that_doesnt_exist.txt/',
    '/missingfolder/./././file_that_doesnt_exist.txt/',
])
def test_rm_file_doesnt_exist(afc: AfcService, path):
    with pytest.raises(AfcFileNotFoundError) as e:
        afc.rm(path)
    assert e.value.status == afc_error_t.OBJECT_NOT_FOUND


def test_get_device_info(afc: AfcService):
    device_info = afc.get_device_info()
    assert device_info['Model'].startswith('iPhone')
    assert int(device_info['FSTotalBytes']) > int(device_info['FSFreeBytes'])


def test_listdir(afc):
    filenames = afc.listdir('/')
    assert 'DCIM' in filenames
    assert 'Downloads' in filenames
    assert 'Books' in filenames


@pytest.mark.parametrize('path', [
    'missing_folder',
    'missingfolder/missing_folder',
    'missingfolder/missing_folder/',
    '/missingfolder/missing_folder',
    '/missingfolder/missing_folder/',
])
def test_listdir_folder_doesnt_exist(afc: AfcService, path):
    with pytest.raises(AfcFileNotFoundError) as e:
        afc.listdir(path)
    assert e.value.status == afc_error_t.OBJECT_NOT_FOUND


def test_listdir_file(afc: AfcService):
    afc.set_file_contents(TEST_FILENAME, b'data')
    try:
        with pytest.raises(AfcException) as e:
            afc.listdir(TEST_FILENAME)
    finally:
        afc.rm(TEST_FILENAME)
    assert e.value.status == afc_error_t.READ_ERROR


@pytest.mark.parametrize('path', [
    'test_dir_a/test_dir_b/test_dir_c/test_dir_d',
    'test_dir_a/test_dir_b/test_dir_c/test_dir_d/',
    '/test_dir_a/test_dir_b/test_dir_c/test_dir_d',
    '/test_dir_a/test_dir_b/test_dir_c/test_dir_d/',
    '/test_dir_a/./../test_dir_a/test_dir_b/../test_dir_b/test_dir_c/',
])
def test_makedirs_and_rm_dir(afc: AfcService, path):
    assert not afc.exists(path)
    afc.makedirs(path)
    assert afc.exists(path)
    afc.rm(pathlib.PosixPath(path.lstrip('/')).parts[0])
    assert not afc.exists(path)


def test_makedirs_file(afc: AfcService):
    afc.set_file_contents(TEST_FILENAME, b'data')
    impossible_path = f'{TEST_FILENAME}/sub_folder'
    try:
        with pytest.raises(AfcException) as e:
            afc.makedirs(impossible_path)
        assert not afc.exists(impossible_path)
    finally:
        afc.rm(TEST_FILENAME)
    assert e.value.status == afc_error_t.OBJECT_EXISTS


def test_isdir_file(afc: AfcService):
    afc.set_file_contents(TEST_FILENAME, b'data')
    assert not afc.isdir(TEST_FILENAME)
    afc.rm(TEST_FILENAME)


@pytest.mark.parametrize('path', [
    TEST_FOLDER_NAME,
    f'{TEST_FOLDER_NAME}/',
    f'/{TEST_FOLDER_NAME}',
    f'/{TEST_FOLDER_NAME}/',
    f'/{TEST_FOLDER_NAME}/./../{TEST_FOLDER_NAME}/',
])
def test_isdir_folder(afc: AfcService, path):
    afc.makedirs(TEST_FILENAME)
    assert afc.isdir(TEST_FILENAME)
    afc.rm(TEST_FILENAME)


def test_isdir_missing_path(afc: AfcService):
    with pytest.raises(AfcFileNotFoundError):
        afc.isdir('folder_that_doesnt_exist')


def test_isdir_missing_path_inside_a_file(afc: AfcService):
    afc.set_file_contents(TEST_FILENAME, b'data')
    impossible_path = f'{TEST_FILENAME}/sub_folder'
    try:
        with pytest.raises(AfcFileNotFoundError):
            afc.isdir(impossible_path)
    finally:
        afc.rm(TEST_FILENAME)


def test_stat_file(afc: AfcService):
    data = b'data'
    timestamp = datetime.fromtimestamp(afc.lockdown.get_value(key='TimeIntervalSince1970'))
    timestamp = timestamp.replace(microsecond=0)  # stat resolution might not include microseconds
    afc.set_file_contents(TEST_FILENAME, data)
    stat = afc.stat(TEST_FILENAME)
    afc.rm(TEST_FILENAME)
    assert stat['st_size'] == len(data)
    assert stat['st_ifmt'] == 'S_IFREG'
    assert stat['st_mtime'] >= timestamp


def test_stat_folder(afc: AfcService):
    timestamp = datetime.fromtimestamp(afc.lockdown.get_value(key='TimeIntervalSince1970'))
    timestamp = timestamp.replace(microsecond=0)  # stat resolution might not include microseconds
    afc.makedirs(TEST_FOLDER_NAME)
    stat = afc.stat(TEST_FOLDER_NAME)
    afc.rm(TEST_FOLDER_NAME)
    assert stat['st_size'] in (64, 68)
    assert stat['st_ifmt'] == 'S_IFDIR'
    assert stat['st_mtime'] >= timestamp


@pytest.mark.parametrize('path', [
    'missing_file',
    '/missing_file',
    'missing_folder/',
    '/missing_folder/',
    'missingfolder/missing_file',
    'missingfolder/missing_folder/',
    '/missingfolder/missing_file',
    '/missingfolder/missing_folder/',
])
def test_stat_doesnt_exist(afc: AfcService, path):
    with pytest.raises(AfcFileNotFoundError):
        afc.stat(path)


def test_stat_missing_path_inside_a_file(afc: AfcService):
    afc.set_file_contents(TEST_FILENAME, b'data')
    impossible_path = f'{TEST_FILENAME}/sub_folder'
    try:
        with pytest.raises(AfcFileNotFoundError):
            afc.stat(impossible_path)
    finally:
        afc.rm(TEST_FILENAME)


def test_fopen_missing_file(afc: AfcService):
    with pytest.raises(AfcFileNotFoundError):
        afc.fopen('file_that_doesnt_exist')


def test_fclose_not_opened(afc: AfcService):
    with pytest.raises(AfcException) as e:
        afc.fclose(77)
    assert e.value.status == afc_error_t.INVALID_ARG


def test_rename(afc: AfcService):
    afc.set_file_contents('source.txt', b'data')
    afc.rename('source.txt', 'dest.txt')
    try:
        assert afc.get_file_contents('dest.txt') == b'data'
    finally:
        afc.rm('dest.txt')
    with pytest.raises(AfcFileNotFoundError):
        afc.get_file_contents('source.txt')


def test_rename_between_folders(afc: AfcService):
    afc.makedirs('dir_a/dir_b')
    source = 'dir_a/dir_b/source.txt'
    dest = 'dir_a/source.txt'
    afc.set_file_contents(source, b'data')
    afc.rename(source, dest)
    try:
        assert afc.get_file_contents(dest) == b'data'
        with pytest.raises(AfcFileNotFoundError):
            afc.get_file_contents(source)
    finally:
        afc.rm('dir_a')


def test_rename_missing_source(afc: AfcService):
    with pytest.raises(AfcFileNotFoundError):
        afc.rename('source.txt', 'dest.txt')


def test_rename_source_path_inside_a_file(afc: AfcService):
    afc.set_file_contents(TEST_FILENAME, b'data')
    impossible_path = f'{TEST_FILENAME}/source.txt'
    try:
        with pytest.raises(AfcFileNotFoundError):
            afc.rename(impossible_path, 'dest.txt')
    finally:
        afc.rm(TEST_FILENAME)


def test_rename_dest_path_inside_a_file(afc: AfcService):
    afc.set_file_contents(TEST_FILENAME, b'data')
    source = 'source.txt'
    afc.set_file_contents(source, b'data')
    impossible_path = f'{TEST_FILENAME}/dest.txt'
    try:
        with pytest.raises(AfcException):
            afc.rename(source, impossible_path)
    finally:
        afc.rm(TEST_FILENAME)
        afc.rm(source)


def test_rename_to_self(afc: AfcService):
    data = b'data'
    afc.set_file_contents(TEST_FILENAME, data)
    afc.rename(TEST_FILENAME, TEST_FILENAME)
    assert afc.get_file_contents(TEST_FILENAME) == data
    afc.rm(TEST_FILENAME)


def test_fread_more_than_file_size(afc: AfcService):
    data = b'data'
    afc.set_file_contents(TEST_FILENAME, data)
    h = afc.fopen(TEST_FILENAME)
    read_data = afc.fread(h, len(data) + 2)
    afc.fclose(h)
    afc.rm(TEST_FILENAME)
    assert read_data == data


def test_fread_not_opened(afc: AfcService):
    with pytest.raises(AfcException) as e:
        afc.fread(77, 4)
    assert e.value.status == afc_error_t.INVALID_ARG


def test_fwrite_not_opened(afc: AfcService):
    with pytest.raises(AfcException) as e:
        afc.fwrite(77, b'asdasd')
    assert e.value.status == afc_error_t.INVALID_ARG


def test_file_read_write(afc: AfcService):
    body = b'data'

    afc.set_file_contents(TEST_FILENAME, body)
    try:
        assert afc.get_file_contents(TEST_FILENAME) == body
    finally:
        afc.rm(TEST_FILENAME)


def test_get_file_contents_missing_file(afc: AfcService):
    with pytest.raises(AfcFileNotFoundError):
        afc.get_file_contents('missing_file')


def test_dirlist(afc: AfcService):
    afc.makedirs('test_a/test_b/test_c/test_d')
    try:
        assert list(afc.dirlist('/', 0)) == ['/']
        dirlist = list(afc.dirlist('/', 2))
        assert '/' in dirlist
        assert '/test_a' in dirlist
        assert '/test_a/test_b' in dirlist
        assert '/test_a/test_b/test_c' not in dirlist
        assert list(afc.dirlist('test_a', 0)) == ['test_a']
        dirlist = list(afc.dirlist('test_a', 2))
        assert 'test_a' in dirlist
        assert 'test_a/test_b' in dirlist
        assert 'test_a/test_b/test_c' in dirlist
        assert 'test_a/test_b/test_c/test_d' not in dirlist
    finally:
        afc.rm('test_a')


def test_push_pull_bigger_than_max_chunk(afc: AfcService):
    contents = b'x' * MAXIMUM_READ_SIZE * 2
    afc.set_file_contents('test', contents)
    assert contents == afc.get_file_contents('test')
    afc.rm('test')



import logging
import os
import posixpath
import re
from typing import List, Optional, TextIO

import click

from pymobiledevice3.cli.cli_common import Command, get_last_used_terminal_formatting, user_requested_colored_output
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.os_trace import OsTraceService, SyslogLogLevel
from pymobiledevice3.services.syslog import SyslogService

logger = logging.getLogger(__name__)


@click.group()
def cli() -> None:
    pass


@cli.group()
def syslog() -> None:
    """ Watch syslog messages """
    pass


@syslog.command('live-old', cls=Command)
def syslog_live_old(service_provider: LockdownClient):
    """ view live syslog lines in raw bytes form from old relay """
    for line in SyslogService(service_provider=service_provider).watch():
        print(line)


def format_line(color, pid, syslog_entry, include_label):
    log_level_colors = {
        SyslogLogLevel.NOTICE.name: 'white',
        SyslogLogLevel.INFO.name: 'white',
        SyslogLogLevel.DEBUG.name: 'green',
        SyslogLogLevel.ERROR.name: 'red',
        SyslogLogLevel.FAULT.name: 'red',
        SyslogLogLevel.USER_ACTION.name: 'white',
    }

    syslog_pid = syslog_entry.pid
    timestamp = syslog_entry.timestamp
    level = syslog_entry.level.name
    filename = syslog_entry.filename
    image_name = posixpath.basename(syslog_entry.image_name)
    message = syslog_entry.message
    process_name = posixpath.basename(filename)
    label = ''

    if (pid != -1) and (syslog_pid != pid):
        return None

    if syslog_entry.label is not None:
        label = f'[{syslog_entry.label.subsystem}][{syslog_entry.label.category}]'

    if color:
        timestamp = click.style(str(timestamp), 'green')
        process_name = click.style(process_name, 'magenta')
        if len(image_name) > 0:
            image_name = click.style(image_name, 'magenta')
        syslog_pid = click.style(syslog_pid, 'cyan')
        log_level_color = log_level_colors[level]
        level = click.style(level, log_level_color)
        label = click.style(label, 'cyan')
        message = click.style(message, log_level_color)

    line_format = '{timestamp} {process_name}{{{image_name}}}[{pid}] <{level}>: {message}'

    if include_label:
        line_format += f' {label}'

    line = line_format.format(timestamp=timestamp, process_name=process_name, image_name=image_name, pid=syslog_pid,
                              level=level, message=message)

    return line


def syslog_live(
        service_provider: LockdownServiceProvider, out: Optional[TextIO], pid: Optional[int],
        process_name: Optional[str], match: List[str], match_insensitive: List[str], include_label: bool,
        regex: List[str], insensitive_regex: List[str]) -> None:
    match_regex = [re.compile(f'.*({r}).*', re.DOTALL) for r in regex]
    match_regex += [re.compile(f'.*({r}).*', re.IGNORECASE | re.DOTALL) for r in insensitive_regex]

    def replace(m):
        if len(m.groups()):
            return line.replace(m.group(1), click.style(m.group(1), bold=True, underline=True))
        return None

    for syslog_entry in OsTraceService(lockdown=service_provider).syslog(pid=pid):
        if process_name:
            if posixpath.basename(syslog_entry.filename) != process_name:
                continue

        line_no_style = format_line(False, pid, syslog_entry, include_label)
        line = format_line(user_requested_colored_output(), pid, syslog_entry, include_label)

        skip = False

        if match is not None:
            for m in match:
                match_line = line
                if m not in match_line:
                    skip = True
                    break
                else:
                    if user_requested_colored_output():
                        match_line = match_line.replace(m, click.style(m, bold=True, underline=True))
                        line = match_line

        if match_insensitive is not None:
            for m in match_insensitive:
                m = m.lower()
                if m not in line.lower():
                    skip = True
                    break
                else:
                    if user_requested_colored_output():
                        start = line.lower().index(m)
                        end = start + len(m)
                        last_color_formatting = get_last_used_terminal_formatting(line[:start])
                        line = line[:start] + click.style(line[start:end], bold=True,
                                                          underline=True) + last_color_formatting + line[end:]

        if match_regex:
            skip = True
            for r in match_regex:
                if not r.findall(line_no_style):
                    continue
                line = re.sub(r, replace, line)
                skip = False

        if skip:
            continue

        print(line, flush=True)

        if out:
            if user_requested_colored_output():
                print(line, file=out, flush=True)
            else:
                print(line_no_style, file=out, flush=True)


@syslog.command('live', cls=Command)
@click.option('-o', '--out', type=click.File('wt'), help='log file')
@click.option('--pid', type=click.INT, default=-1, help='pid to filter. -1 for all')
@click.option('-pn', '--process-name', help='process name to filter')
@click.option('-m', '--match', multiple=True, help='match expression')
@click.option('-mi', '--match-insensitive', multiple=True, help='insensitive match expression')
@click.option('include_label', '--label', is_flag=True, help='should include label')
@click.option('-e', '--regex', multiple=True, help='filter only lines matching given regex')
@click.option('-ei', '--insensitive-regex', multiple=True, help='filter only lines matching given regex (insensitive)')
def cli_syslog_live(
        service_provider: LockdownServiceProvider, out: Optional[TextIO], pid: Optional[int],
        process_name: Optional[str], match: List[str], match_insensitive: List[str], include_label: bool,
        regex: List[str], insensitive_regex: List[str]) -> None:
    """ view live syslog lines """

    syslog_live(service_provider, out, pid, process_name, match, match_insensitive, include_label, regex,
                insensitive_regex)


@syslog.command('collect', cls=Command)
@click.argument('out', type=click.Path(exists=False, dir_okay=True, file_okay=False))
@click.option('--size-limit', type=click.INT)
@click.option('--age-limit', type=click.INT)
@click.option('--start-time', type=click.INT)
def syslog_collect(service_provider: LockdownClient, out, size_limit, age_limit, start_time):
    """
    Collect the system logs into a .logarchive that can be viewed later with tools such as log or Console.
    If the filename doesn't exist, system_logs.logarchive will be created in the given directory.
    """
    if not os.path.exists(out):
        os.makedirs(out)

    if not out.endswith('.logarchive'):
        logger.warning('given out path doesn\'t end with a .logarchive - consider renaming to be able to view '
                       'the file with the likes of the Console.app and the `log show` utilities')

    OsTraceService(lockdown=service_provider).collect(out, size_limit=size_limit, age_limit=age_limit,
                                                      start_time=start_time)



import datetime
import logging
from abc import abstractmethod
from typing import Optional

from pymobiledevice3.exceptions import StartServiceError
from pymobiledevice3.service_connection import ServiceConnection


class LockdownServiceProvider:
    def __init__(self):
        self.udid: Optional[str] = None
        self.product_type: Optional[str] = None

    @property
    @abstractmethod
    def product_version(self) -> str:
        pass

    @property
    @abstractmethod
    def ecid(self) -> int:
        pass

    @property
    @abstractmethod
    def developer_mode_status(self) -> bool:
        pass

    @property
    @abstractmethod
    def date(self) -> datetime.datetime:
        pass

    @abstractmethod
    def set_language(self, language: str) -> None:
        pass

    @abstractmethod
    def start_lockdown_service(self, name: str, include_escrow_bag: bool = False) -> ServiceConnection:
        pass

    @abstractmethod
    async def aio_start_lockdown_service(
            self, name: str, include_escrow_bag: bool = False) -> ServiceConnection:
        pass

    def start_lockdown_developer_service(
            self, name: str, include_escrow_bag: bool = False) -> ServiceConnection:
        try:
            return self.start_lockdown_service(name, include_escrow_bag=include_escrow_bag)
        except StartServiceError:
            logging.getLogger(self.__module__).error(
                'Failed to connect to required service. Make sure DeveloperDiskImage.dmg has been mounted. '
                'You can do so using: pymobiledevice3 mounter mount'
            )
            raise



import glob
import shutil
import time

import pytest

from pymobiledevice3.lockdown import create_using_usbmux
from pymobiledevice3.services.crash_reports import CrashReportsManager

BASENAME = '__pymobiledevice3_tests'
PATH_COMPONENT = f'/{BASENAME}'


@pytest.fixture(scope='function')
def crash_manager(lockdown):
    with CrashReportsManager(lockdown) as crash_manager:
        yield crash_manager


@pytest.fixture(scope='module')
def delete_test_dir():
    yield
    with create_using_usbmux() as lockdown_client:
        with CrashReportsManager(lockdown_client) as crash_manager:
            crash_manager.afc.rm(BASENAME)


def test_ls_default(crash_manager, delete_test_dir):
    crash_manager.afc.makedirs(PATH_COMPONENT)
    assert PATH_COMPONENT in crash_manager.ls()


def test_ls_path(crash_manager, delete_test_dir):
    crash_manager.afc.makedirs(PATH_COMPONENT * 2)
    assert (PATH_COMPONENT * 2) in crash_manager.ls(path=PATH_COMPONENT)


@pytest.mark.parametrize('depth', [2, 3, 4])
def test_ls_depth(crash_manager, delete_test_dir, depth):
    path = PATH_COMPONENT * depth
    path_list = [PATH_COMPONENT * i for i in range(1, depth + 1)]
    crash_manager.afc.makedirs(path)
    crash_list = crash_manager.ls(depth=depth)
    for item in path_list:
        assert item in crash_list


def test_ls_depth_minus_one(crash_manager, delete_test_dir):
    path_list = [PATH_COMPONENT, PATH_COMPONENT * 2, PATH_COMPONENT * 3]
    crash_manager.afc.makedirs(path_list[-1])
    crash_list = crash_manager.ls(depth=-1)
    for path in path_list:
        assert path in crash_list


def test_clear(crash_manager, delete_test_dir):
    crash_manager.afc.makedirs(PATH_COMPONENT)
    # true indication device time we can assure that every other file should create after it
    test_dir_birth_time = crash_manager.afc.stat(PATH_COMPONENT)['st_birthtime']
    crash_manager.clear()
    crash_dirlist = crash_manager.ls(depth=-1)
    assert PATH_COMPONENT not in crash_dirlist
    for path in crash_dirlist:
        if crash_manager.APPSTORED_PATH != path:
            assert crash_manager.afc.stat(path)['st_birthtime'] > test_dir_birth_time


def test_pull(crash_manager, delete_test_dir):
    crash_manager.afc.makedirs(PATH_COMPONENT)
    dir_list = crash_manager.ls(depth=-1)
    crash_manager.pull(BASENAME)
    pulled_list = [file[len(BASENAME):] for file in glob.glob(f'{BASENAME}/**', recursive=True)][1:]  # ignore root path
    assert sorted(dir_list) == sorted(pulled_list)
    shutil.rmtree(BASENAME)


@pytest.mark.parametrize(
    ('end_time', 'return_value'),
    ((-1, True), (0, True), (time.monotonic() + 1000, False), (None, False))
)
def test_check_timeout(crash_manager, end_time, return_value):
    assert crash_manager._check_timeout(end_time) is return_value



import asyncio
import dataclasses
import sys
from socket import AF_INET, AF_INET6, inet_ntop
from typing import List, Mapping, Optional

from ifaddr import get_adapters
from zeroconf import IPVersion, ServiceListener, ServiceStateChange, Zeroconf
from zeroconf.asyncio import AsyncServiceBrowser, AsyncServiceInfo, AsyncZeroconf

from pymobiledevice3.exceptions import FeatureNotSupportedError
from pymobiledevice3.osu.os_utils import get_os_utils

REMOTEPAIRING_SERVICE_NAMES = ['_remotepairing._tcp.local.']
REMOTEPAIRING_MANUAL_PAIRING_SERVICE_NAMES = ['_remotepairing-manual-pairing._tcp.local.']
MOBDEV2_SERVICE_NAMES = ['_apple-mobdev2._tcp.local.']
REMOTED_SERVICE_NAMES = ['_remoted._tcp.local.']
OSUTILS = get_os_utils()
DEFAULT_BONJOUR_TIMEOUT = OSUTILS.bonjour_timeout


@dataclasses.dataclass
class BonjourAnswer:
    name: str
    properties: Mapping[bytes, bytes]
    ips: List[str]
    port: int


class BonjourListener(ServiceListener):
    def __init__(self, ip: str):
        super().__init__()
        self.name: Optional[str] = None
        self.properties: Mapping[bytes, bytes] = {}
        self.ip = ip
        self.port: Optional[int] = None
        self.addresses: List[str] = []
        self.queue: asyncio.Queue = asyncio.Queue()
        self.querying_task: Optional[asyncio.Task] = asyncio.create_task(self.query_addresses())

    def async_on_service_state_change(
            self, zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange) -> None:
        self.queue.put_nowait((zeroconf, service_type, name, state_change))

    async def query_addresses(self) -> None:
        zeroconf, service_type, name, state_change = await self.queue.get()
        self.name = name
        service_info = AsyncServiceInfo(service_type, name)
        await service_info.async_request(zeroconf, 3000)
        ipv4 = [inet_ntop(AF_INET, address.packed) for address in
                service_info.ip_addresses_by_version(IPVersion.V4Only)]
        ipv6 = []
        if '%' in self.ip:
            ipv6 = [inet_ntop(AF_INET6, address.packed) + '%' + self.ip.split('%')[1] for address in
                    service_info.ip_addresses_by_version(IPVersion.V6Only)]
        self.addresses = ipv4 + ipv6
        self.properties = service_info.properties
        self.port = service_info.port

    async def close(self) -> None:
        self.querying_task.cancel()
        try:
            await self.querying_task
        except asyncio.CancelledError:
            pass


@dataclasses.dataclass
class BonjourQuery:
    zc: AsyncZeroconf
    service_browser: AsyncServiceBrowser
    listener: BonjourListener


def query_bonjour(service_names: List[str], ip: str) -> BonjourQuery:
    aiozc = AsyncZeroconf(interfaces=[ip])
    listener = BonjourListener(ip)
    service_browser = AsyncServiceBrowser(aiozc.zeroconf, service_names,
                                          handlers=[listener.async_on_service_state_change])
    return BonjourQuery(aiozc, service_browser, listener)


async def browse(service_names: List[str], ips: List[str], timeout: float = DEFAULT_BONJOUR_TIMEOUT) \
        -> List[BonjourAnswer]:
    try:
        bonjour_queries = [query_bonjour(service_names, adapter) for adapter in ips]
    except ValueError as e:
        if sys.version_info.major == 3 and sys.version_info.minor == 8:
            raise FeatureNotSupportedError('python3.8', 'IPv6 link-local bonjour browse') from e
        else:
            raise e
    answers = []
    await asyncio.sleep(timeout)
    for bonjour_query in bonjour_queries:
        if bonjour_query.listener.addresses:
            answer = BonjourAnswer(
                bonjour_query.listener.name, bonjour_query.listener.properties, bonjour_query.listener.addresses,
                bonjour_query.listener.port)
            if answer not in answers:
                answers.append(answer)
        await bonjour_query.listener.close()
        await bonjour_query.service_browser.async_cancel()
        await bonjour_query.zc.async_close()
    return answers


async def browse_ipv6(service_names: List[str], timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> List[BonjourAnswer]:
    return await browse(service_names, OSUTILS.get_ipv6_ips(), timeout=timeout)


def get_ipv4_addresses() -> List[str]:
    ips = []
    for adapter in get_adapters():
        if adapter.nice_name.startswith('tun'):
            # skip browsing on already established tunnels
            continue
        for ip in adapter.ips:
            if ip.ip == '127.0.0.1':
                continue
            if not ip.is_IPv4:
                continue
            ips.append(ip.ip)
    return ips


async def browse_ipv4(service_names: List[str], timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> List[BonjourAnswer]:
    return await browse(service_names, get_ipv4_addresses(), timeout=timeout)


async def browse_remoted(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> List[BonjourAnswer]:
    return await browse_ipv6(REMOTED_SERVICE_NAMES, timeout=timeout)


async def browse_mobdev2(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> List[BonjourAnswer]:
    return await browse(MOBDEV2_SERVICE_NAMES, get_ipv4_addresses() + OSUTILS.get_ipv6_ips(), timeout=timeout)


async def browse_remotepairing(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> List[BonjourAnswer]:
    return await browse_ipv4(REMOTEPAIRING_SERVICE_NAMES, timeout=timeout)


async def browse_remotepairing_manual_pairing(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> List[BonjourAnswer]:
    return await browse_ipv4(REMOTEPAIRING_MANUAL_PAIRING_SERVICE_NAMES, timeout=timeout)



import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.companion import CompanionProxyService


@click.group()
def cli() -> None:
    pass


@cli.group()
def companion() -> None:
    """ List paired "companion" devices """
    pass


@companion.command('list', cls=Command)
def companion_list(service_provider: LockdownClient):
    """ list all paired companion devices """
    print_json(CompanionProxyService(service_provider).list(), default=lambda x: '<non-serializable>')



import logging

import pytest
import pytest_asyncio

from pymobiledevice3.exceptions import DeviceNotFoundError, InvalidServiceError
from pymobiledevice3.lockdown import LockdownClient, create_using_usbmux
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.tunneld import async_get_tunneld_devices

logging.getLogger('quic').disabled = True
logging.getLogger('asyncio').disabled = True
logging.getLogger('zeroconf').disabled = True
logging.getLogger('parso.cache').disabled = True
logging.getLogger('parso.cache.pickle').disabled = True
logging.getLogger('parso.python.diff').disabled = True
logging.getLogger('humanfriendly.prompts').disabled = True
logging.getLogger('blib2to3.pgen2.driver').disabled = True
logging.getLogger('urllib3.connectionpool').disabled = True


def pytest_addoption(parser):
    parser.addoption('--rsd', default=None, type=str, nargs=2, action='store')
    parser.addoption('--tunnel', default=None, type=str, action='store')


@pytest.fixture(scope='function')
def rsd_option(request):
    """
    Get --rsd option
    """
    return request.config.getoption('--rsd')


@pytest.fixture(scope='function')
def tunnel_option(request):
    """
    Get --tunnel option
    """
    return request.config.getoption('--tunnel')


@pytest_asyncio.fixture(scope='function')
async def service_provider(rsd_option, tunnel_option) -> LockdownServiceProvider:
    """
    Creates a new LockdownServiceProvider client for each test.
    """
    if rsd_option is not None:
        async with RemoteServiceDiscoveryService(rsd_option) as rsd:
            yield rsd
    elif tunnel_option is not None:
        rsds = await async_get_tunneld_devices()
        try:
            if tunnel_option == '':
                yield rsds[0]
            else:
                yield [rsd for rsd in rsds if rsd.udid == tunnel_option][0]
            for rsd in rsds:
                await rsd.close()
        except IndexError:
            raise DeviceNotFoundError(tunnel_option)
    else:
        with create_using_usbmux() as client:
            yield client


@pytest.fixture(scope='function')
def dvt(service_provider) -> DvtSecureSocketProxyService:
    """
    Creates a new DvtSecureSocketProxyService client for each test.
    """
    try:
        with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
            yield dvt
    except InvalidServiceError:
        pytest.skip('Skipping DVT-based test since the service isn\'t accessible')


@pytest.fixture(scope='function')
def lockdown(request) -> LockdownClient:
    """
    Creates a new lockdown client for each test.
    """
    with create_using_usbmux() as client:
        yield client



from datetime import datetime
from typing import IO, Optional

import click
from pygments import formatters, highlight, lexers

from pymobiledevice3.cli.cli_common import Command, print_hex, user_requested_colored_output
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.pcapd import PcapdService


@click.group()
def cli() -> None:
    pass


def print_packet_header(packet, color: bool):
    date = datetime.fromtimestamp(packet.seconds + (packet.microseconds / 1000000))
    data = (
        f'{date}: '
        f'Process {packet.comm} ({packet.pid}), '
        f'Interface: {packet.interface_name} ({packet.interface_type.name}), '
        f'Family: {packet.protocol_family.name}'
    )
    if not color:
        print(data)
    else:
        print(highlight(data, lexers.HspecLexer(), formatters.TerminalTrueColorFormatter(style='native')), end='')


def print_packet(packet, color: Optional[bool] = None):
    """ Return the packet, so it can be chained in a generator """
    if color is None:
        color = user_requested_colored_output()
    print_packet_header(packet, color)
    print_hex(packet.data, color)
    return packet


@cli.command(cls=Command)
@click.argument('out', type=click.File('wb'), required=False)
@click.option('-c', '--count', type=click.INT, default=-1, help='Number of packets to sniff. Omit to endless sniff.')
@click.option('--process', default=None, help='Process to filter. Omit for all.')
@click.option('-i', '--interface', default=None, help='Interface name to filter. Omit for all.')
def pcap(service_provider: LockdownServiceProvider, out: Optional[IO], count: int, process: Optional[str],
         interface: Optional[str]) -> None:
    """ Sniff device traffic """
    service = PcapdService(lockdown=service_provider)
    packets_generator = service.watch(packets_count=count, process=process, interface_name=interface)

    if out is not None:
        packets_generator_with_print = map(lambda p: print_packet(p), packets_generator)
        service.write_to_pcap(out, packets_generator_with_print)
        return

    for packet in packets_generator:
        print_packet(packet)



import time

import pytest

from pymobiledevice3.exceptions import DvtDirListError, UnrecognizedSelectorError
from pymobiledevice3.services.dvt.instruments.application_listing import ApplicationListing
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl


def get_process_data(dvt, name: str):
    processes = DeviceInfo(dvt).proclist()
    return [process for process in processes if process['name'] == name][0]


def test_ls(dvt):
    """
    Test listing a directory.
    """
    ls = set(DeviceInfo(dvt).ls('/'))
    assert {'usr', 'bin', 'etc', 'var', 'private', 'Applications', 'Developer'} <= ls


def test_ls_failure(dvt):
    """
    Test listing a directory.
    """
    with pytest.raises(DvtDirListError):
        DeviceInfo(dvt).ls('Directory that does not exist')


def test_proclist(dvt):
    """
    Test listing processes.
    """
    lockdownd = get_process_data(dvt, 'lockdownd')
    assert lockdownd['realAppName'] == '/usr/libexec/lockdownd'
    assert not lockdownd['isApplication']


def test_applist(dvt):
    """
    Test listing applications.
    """
    apps = ApplicationListing(dvt).applist()
    safari = [app for app in apps if app['DisplayName'] == 'StocksWidget'][0]
    assert safari['CFBundleIdentifier'] == 'com.apple.stocks.widget'
    assert safari['Restricted'] == 1
    assert safari['Type'] == 'PluginKit'


def test_kill(dvt):
    """
    Test killing a process.
    """
    aggregated = get_process_data(dvt, 'SpringBoard')
    ProcessControl(dvt).kill(aggregated['pid'])
    # give the os some time to start the process again
    time.sleep(3)
    aggregated_after_kill = get_process_data(dvt, 'SpringBoard')
    if 'startDate' in aggregated:
        assert aggregated['startDate'] < aggregated_after_kill['startDate']


def test_launch(dvt):
    """
    Test launching a process.
    """
    pid = ProcessControl(dvt).launch('com.apple.mobilesafari')
    assert pid
    for process in DeviceInfo(dvt).proclist():
        if pid == process['pid']:
            assert process['name'] == 'MobileSafari'


def test_system_information(dvt):
    """
    Test getting system information.
    """
    try:
        system_info = DeviceInfo(dvt).system_information()
    except UnrecognizedSelectorError:
        pytest.skip('device doesn\'t support this method')
    assert '_deviceDescription' in system_info and system_info['_deviceDescription'].startswith('Build Version')


def test_hardware_information(dvt):
    """
    Test getting hardware information.
    """
    hardware_info = DeviceInfo(dvt).hardware_information()
    assert hardware_info['numberOfCpus'] > 0


def test_network_information(dvt):
    """
    Test getting network information.
    """
    network_info = DeviceInfo(dvt).network_information()
    assert network_info['lo0'] == 'Loopback'



import pytest

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.web_protocol.driver import WebDriver
from pymobiledevice3.services.webinspector import SAFARI, WebinspectorService


@pytest.fixture
def webdriver(lockdown: LockdownClient):
    inspector = WebinspectorService(lockdown=lockdown)
    inspector.connect()
    safari = inspector.open_app(SAFARI)
    inspector.flush_input(1)
    session = inspector.automation_session(safari)
    driver = WebDriver(session)
    driver.start_session()
    try:
        yield driver
    finally:
        inspector.close()
        inspector.loop.close()



from pymobiledevice3.services.web_protocol.driver import By
from tests.services.test_web_protocol.common import LINK_HTML


def test_back(webdriver):
    webdriver.get('https://www.google.com')
    webdriver.get('https://www.github.com')
    webdriver.back()
    assert webdriver.current_url.rstrip('/') == 'https://www.google.com'


def test_current_url(webdriver):
    assert not webdriver.current_url
    url = 'https://www.google.com'
    webdriver.get(url)
    assert webdriver.current_url.rstrip('/') == url


def test_forward(webdriver):
    webdriver.get('https://www.google.com')
    webdriver.get('https://www.github.com')
    webdriver.back()
    webdriver.forward()
    assert webdriver.current_url.rstrip('/') == 'https://www.github.com'


def test_find_element(webdriver):
    webdriver.execute_script(f'''document.getElementsByTagName('body')[0].innerHTML = '{LINK_HTML}'; ''')
    by_id = webdriver.find_element(By.ID, 'id_of_link')
    by_xpath = webdriver.find_element(By.XPATH, '/html/body/a')
    by_link_text = webdriver.find_element(By.LINK_TEXT, 'the text obviously')
    by_partial_link_text = webdriver.find_element(By.PARTIAL_LINK_TEXT, 'obviously')
    by_name = webdriver.find_element(By.NAME, 'name_of_link')
    by_tag = webdriver.find_element(By.TAG_NAME, 'a')
    by_class = webdriver.find_element(By.CLASS_NAME, 'link-class')
    by_css = webdriver.find_element(By.CSS_SELECTOR, '.link-class')
    assert by_id == by_xpath == by_link_text == by_partial_link_text == by_name == by_tag == by_class == by_css
    assert by_id is not None



from pathlib import Path

from pymobiledevice3.osu.os_utils import get_os_utils

_HOMEFOLDER = get_os_utils().get_homedir() / '.pymobiledevice3'


def get_home_folder() -> Path:
    _HOMEFOLDER.mkdir(exist_ok=True, parents=True)
    return _HOMEFOLDER



from pathlib import Path

import pytest
from packaging.version import Version

from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.services.dtfetchsymbols import DtFetchSymbols
from pymobiledevice3.services.remote_fetch_symbols import RemoteFetchSymbolsService


@pytest.mark.asyncio
async def test_fetch_symbols_list(service_provider):
    """
    Test listing of device symbol files
    """
    if Version(service_provider.product_version) < Version('17.0'):
        DtFetchSymbols(service_provider).list_files()
    else:
        if not isinstance(service_provider, RemoteServiceDiscoveryService):
            pytest.skip('requires RemoteServiceDiscoveryService')

        async with RemoteFetchSymbolsService(service_provider) as fetch_symbols:
            assert len(await fetch_symbols.get_dsc_file_list()) > 0


@pytest.mark.asyncio
async def test_fetch_symbols_download(service_provider, tmp_path):
    """
    Test download of device symbol files
    """
    if Version(service_provider.product_version) < Version('17.0'):
        tmp_file = Path(tmp_path) / 'tmp'
        with tmp_file.open('wb') as file:
            DtFetchSymbols(service_provider).get_file(0, file)
    else:
        if not isinstance(service_provider, RemoteServiceDiscoveryService):
            pytest.skip('requires RemoteServiceDiscoveryService')

        async with RemoteFetchSymbolsService(service_provider) as fetch_symbols:
            await fetch_symbols.download(tmp_path)



from pymobiledevice3.services.installation_proxy import InstallationProxyService


def test_get_apps(lockdown):
    with InstallationProxyService(lockdown=lockdown) as installation_proxy:
        apps = installation_proxy.get_apps()
        assert len(apps) > 1


def test_get_system_apps(lockdown):
    with InstallationProxyService(lockdown=lockdown) as installation_proxy:
        app_types = set(
            [app['ApplicationType'] for app in installation_proxy.get_apps(application_type='System').values()])
        assert len(app_types) == 1
        assert app_types.pop() == 'System'



#!/usr/bin/env python3
import logging
import time

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider


class HeartbeatService:
    """
    Use to keep an active connection with lockdowd
    """
    SERVICE_NAME = 'com.apple.mobile.heartbeat'
    RSD_SERVICE_NAME = 'com.apple.mobile.heartbeat.shim.remote'

    def __init__(self, lockdown: LockdownServiceProvider):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown

        if isinstance(lockdown, LockdownClient):
            self.service_name = self.SERVICE_NAME
        else:
            self.service_name = self.RSD_SERVICE_NAME

    def start(self, interval: float = None) -> None:
        start = time.time()
        service = self.lockdown.start_lockdown_service(self.service_name)

        while True:
            response = service.recv_plist()
            self.logger.debug(response)

            if interval is not None:
                if time.time() >= start + interval:
                    break

            service.send_plist({'Command': 'Polo'})



import asyncio
import plistlib
from pathlib import Path

import click

from pymobiledevice3.bonjour import DEFAULT_BONJOUR_TIMEOUT, browse_remotepairing, browse_remotepairing_manual_pairing
from pymobiledevice3.cli.cli_common import BaseCommand, print_json
from pymobiledevice3.cli.remote import browse_rsd
from pymobiledevice3.lockdown import get_mobdev2_lockdowns


@click.group()
def cli() -> None:
    pass


@cli.group('bonjour')
def bonjour_cli() -> None:
    """ Browse devices over bonjour """
    pass


async def cli_mobdev2_task(timeout: float, pair_records: str) -> None:
    records = []
    if pair_records is not None:
        for record in Path(pair_records).glob('*.plist'):
            records.append(plistlib.loads(record.read_bytes()))
    output = []
    async for ip, lockdown in get_mobdev2_lockdowns(timeout=timeout):
        short_info = lockdown.short_info
        short_info['ip'] = ip
        output.append(short_info)
    print_json(output)


@bonjour_cli.command('mobdev2', cls=BaseCommand)
@click.option('--timeout', default=DEFAULT_BONJOUR_TIMEOUT, type=click.INT)
@click.option('--pair-records', type=click.Path(dir_okay=True, file_okay=False, exists=True),
              help='pair records to attempt validation with')
def cli_mobdev2(timeout: float, pair_records: str) -> None:
    """ browse for mobdev2 devices over bonjour """
    asyncio.run(cli_mobdev2_task(timeout, pair_records))


async def cli_remotepairing_task(timeout: float) -> None:
    output = []
    for answer in await browse_remotepairing(timeout=timeout):
        for ip in answer.ips:
            output.append({'hostname': ip, 'port': answer.port})
    print_json(output)


@bonjour_cli.command('remotepairing', cls=BaseCommand)
@click.option('--timeout', default=DEFAULT_BONJOUR_TIMEOUT, type=click.FLOAT)
def cli_remotepairing(timeout: float) -> None:
    """ browse for remotepairing devices over bonjour (without attempting pair verification) """
    asyncio.run(cli_remotepairing_task(timeout=timeout))


async def cli_remotepairing_manual_pairing_task(timeout: float) -> None:
    output = []
    for answer in await browse_remotepairing_manual_pairing(timeout=timeout):
        for ip in answer.ips:
            output.append({'hostname': ip, 'port': answer.port, 'name': answer.properties[b'name'].decode()})
    print_json(output)


@bonjour_cli.command('remotepairing-manual-pairing', cls=BaseCommand)
@click.option('--timeout', default=DEFAULT_BONJOUR_TIMEOUT, type=click.FLOAT)
def cli_remotepairing_manual_pairing(timeout: float) -> None:
    """ browse for remotepairing-manual-pairing devices over bonjour """
    asyncio.run(cli_remotepairing_manual_pairing_task(timeout=timeout))


async def cli_browse_rsd() -> None:
    print_json(await browse_rsd())


@bonjour_cli.command('rsd', cls=BaseCommand)
def cli_rsd() -> None:
    """ browse RemoteXPC devices using bonjour """
    asyncio.run(cli_browse_rsd(), debug=True)



import threading
from socket import socket

import pytest

from pymobiledevice3.lockdown import SERVICE_PORT, LockdownClient
from pymobiledevice3.tcp_forwarder import UsbmuxTcpForwarder

FREE_PORT = 3582


def attempt_local_connection(port: int):
    client = socket()
    client.connect(('127.0.0.1', port))
    client.close()


@pytest.mark.parametrize('dst_port', [FREE_PORT, SERVICE_PORT])
def test_tcp_forwarder_bad_port(lockdown: LockdownClient, dst_port: int):
    # start forwarder
    listening_event = threading.Event()
    forwarder = UsbmuxTcpForwarder(lockdown.udid, dst_port, FREE_PORT, listening_event=listening_event)
    thread = threading.Thread(target=forwarder.start)
    thread.start()

    # wait for it to actually start listening
    listening_event.wait()
    attempt_local_connection(FREE_PORT)

    # tell it to stop
    forwarder.stop()

    # make sure it stops
    thread.join()



import datetime
import os
import signal
import socket
import struct
from pathlib import Path
from typing import List, Tuple

from ifaddr import get_adapters

from pymobiledevice3.osu.os_utils import DEFAULT_AFTER_IDLE_SEC, DEFAULT_INTERVAL_SEC, DEFAULT_MAX_FAILS, OsUtils
from pymobiledevice3.usbmux import MuxConnection

_DARWIN_TCP_KEEPALIVE = 0x10
_DARWIN_TCP_KEEPINTVL = 0x101
_DARWIN_TCP_KEEPCNT = 0x102


class Posix(OsUtils):
    @property
    def is_admin(self) -> bool:
        return os.geteuid() == 0

    @property
    def usbmux_address(self) -> Tuple[str, int]:
        return MuxConnection.USBMUXD_PIPE, socket.AF_UNIX

    @property
    def bonjour_timeout(self) -> int:
        return 1

    @property
    def access_denied_error(self) -> str:
        return 'This command requires root privileges. Consider retrying with "sudo".'

    def get_ipv6_ips(self) -> List[str]:
        return [f'{adapter.ips[0].ip[0]}%{adapter.nice_name}' for adapter in get_adapters() if
                adapter.ips[0].is_IPv6 and not adapter.nice_name.startswith('tun')]

    def chown_to_non_sudo_if_needed(self, path: Path) -> None:
        if os.getenv('SUDO_UID') is None:
            return
        os.chown(path, int(os.getenv('SUDO_UID')), int(os.getenv('SUDO_GID')))

    def parse_timestamp(self, time_stamp) -> datetime:
        return datetime.datetime.fromtimestamp(time_stamp)

    def wait_return(self):
        print("Press Ctrl+C to send a SIGINT or use 'kill' command to send a SIGTERM")
        signal.sigwait([signal.SIGINT, signal.SIGTERM])


class Darwin(Posix):
    @property
    def pair_record_path(self) -> Path:
        return Path('/var/db/lockdown/')

    @property
    def loopback_header(self) -> bytes:
        return struct.pack('>I', socket.AF_INET6)

    def set_keepalive(self, sock: socket.socket, after_idle_sec: int = DEFAULT_AFTER_IDLE_SEC,
                      interval_sec: int = DEFAULT_INTERVAL_SEC, max_fails: int = DEFAULT_MAX_FAILS) -> None:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        sock.setsockopt(socket.IPPROTO_TCP, _DARWIN_TCP_KEEPALIVE, after_idle_sec)
        sock.setsockopt(socket.IPPROTO_TCP, _DARWIN_TCP_KEEPINTVL, interval_sec)
        sock.setsockopt(socket.IPPROTO_TCP, _DARWIN_TCP_KEEPCNT, max_fails)


class Linux(Posix):
    @property
    def pair_record_path(self) -> Path:
        return Path('/var/lib/lockdown/')

    @property
    def loopback_header(self) -> bytes:
        return b'\x00\x00\x86\xdd'

    def set_keepalive(self, sock: socket.socket, after_idle_sec: int = DEFAULT_AFTER_IDLE_SEC,
                      interval_sec: int = DEFAULT_INTERVAL_SEC, max_fails: int = DEFAULT_MAX_FAILS) -> None:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, after_idle_sec)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, interval_sec)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, max_fails)

    def get_homedir(self) -> Path:
        return Path('~' + os.environ.get('SUDO_USER', '')).expanduser()


class Cygwin(Posix):
    @property
    def usbmux_address(self) -> Tuple[str, int]:
        return MuxConnection.ITUNES_HOST, socket.AF_INET



import asyncio
import traceback
from functools import wraps
from pathlib import Path
from typing import Callable

import requests
from construct import Int8ul, Int16ul, Int32ul, Int64ul, Select
from tqdm import tqdm


def plist_access_path(d, path: tuple, type_=None, required=False):
    for component in path:
        d = d.get(component)
        if d is None:
            break

    if type_ == bool and isinstance(d, str):
        if d.lower() not in ('true', 'false'):
            raise ValueError()
        d = 'true' == d.lower()
    elif type_ is not None and not isinstance(d, type_):
        # wrong type
        d = None

    if d is None and required:
        raise KeyError(f'path: {path} doesn\'t exist in given plist object')

    return d


def bytes_to_uint(b: bytes):
    return Select(u64=Int64ul, u32=Int32ul, u16=Int16ul, u8=Int8ul).parse(b)


def try_decode(s: bytes):
    try:
        return s.decode('utf8')
    except UnicodeDecodeError:
        return s


def asyncio_print_traceback(f: Callable):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except Exception as e:  # noqa: E72
            if not isinstance(e, asyncio.CancelledError):
                traceback.print_exc()
            raise

    return wrapper


def get_asyncio_loop() -> asyncio.AbstractEventLoop:
    try:
        loop = asyncio.get_running_loop()
        if loop.is_closed():
            raise RuntimeError('The existing loop is closed.')
    except RuntimeError:
        # This happens when there is no current event loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def file_download(url: str, outfile: Path, chunk_size=1024) -> None:
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get('content-length', 0))
    with outfile.open('wb') as file, tqdm(
            desc=outfile.name,
            total=total,
            unit='iB',
            unit_scale=True,
            unit_divisor=1024,
    ) as bar:
        for data in resp.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)



import subprocess
import sys

import pytest
from click.testing import CliRunner

from pymobiledevice3 import __main__

pytestmark = [pytest.mark.cli]


def test_cli_main_interface():
    runner = CliRunner()
    r1 = runner.invoke(__main__.cli, ['--help'])
    assert r1.exit_code == 0

    r2 = runner.invoke(__main__.cli)
    assert r2.exit_code == 0


@pytest.mark.parametrize('group', __main__.CLI_GROUPS.keys())
def test_cli_groups(group):
    runner = CliRunner()
    group_help_result = runner.invoke(__main__.cli, [group, '--help'])
    assert group_help_result.exit_code == 0


@pytest.mark.parametrize('group', __main__.CLI_GROUPS.keys())
def test_cli_from_python_m_flag(group):
    subprocess.run([sys.executable, '-m', 'pymobiledevice3', group, '--help'], check=True)



import logging
import plistlib
import tempfile
from pathlib import Path
from typing import IO, List, Optional

import click

from pymobiledevice3.ca import create_keybag_file
from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.mobile_activation import MobileActivationService
from pymobiledevice3.services.mobile_config import MobileConfigService

logger = logging.getLogger(__name__)


@click.group()
def cli() -> None:
    pass


@cli.group('profile')
def profile_group() -> None:
    """ Managed installed profiles or install SSL certificates """
    pass


@profile_group.command('list', cls=Command)
def profile_list(service_provider: LockdownClient):
    """ List installed profiles """
    print_json(MobileConfigService(lockdown=service_provider).get_profile_list())


@profile_group.command('install', cls=Command)
@click.option('--keybag', type=click.Path(file_okay=True, dir_okay=False, exists=True))
@click.argument('profiles', nargs=-1, type=click.File('rb'))
def profile_install(service_provider: LockdownServiceProvider, keybag: Optional[str], profiles: List[IO]) -> None:
    """
    Install given profiles

    If given a keybag, use that to install the profile silently
    """
    service = MobileConfigService(lockdown=service_provider)
    for profile in profiles:
        logger.info(f'installing {profile.name}')
        if keybag is not None:
            service.install_profile_silent(Path(keybag), profile.read())
        else:
            service.install_profile(profile.read())


@profile_group.command('cloud-configuration', cls=Command)
@click.argument('config', type=click.File('rb'), required=False)
def profile_cloud_configuration(service_provider: LockdownServiceProvider, config: Optional[IO]) -> None:
    """ Get/Set cloud configuration """
    if not config:
        print_json(MobileConfigService(lockdown=service_provider).get_cloud_configuration())
    else:
        config_json = plistlib.load(config)
        logger.info(f'applying cloud configuration {config_json}')
        MobileConfigService(lockdown=service_provider).set_cloud_configuration(config_json)
        logger.info('applied cloud configuration')


@profile_group.command('store', cls=Command)
@click.argument('profiles', nargs=-1, type=click.File('rb'))
def profile_store(service_provider: LockdownServiceProvider, profiles: List[IO]) -> None:
    """ Store a profile """
    service = MobileConfigService(lockdown=service_provider)
    for profile in profiles:
        logger.info(f'storing {profile.name}')
        service.store_profile(profile.read())


@profile_group.command('remove', cls=Command)
@click.argument('name')
def profile_remove(service_provider: LockdownServiceProvider, name: str) -> None:
    """ Remove a profile by its name """
    MobileConfigService(lockdown=service_provider).remove_profile(name)


@profile_group.command('set-wifi-power', cls=Command)
@click.argument('state', type=click.Choice(['on', 'off']), required=False)
def profile_set_wifi_power(service_provider: LockdownServiceProvider, state: str) -> None:
    """ change Wi-Fi power state """
    MobileConfigService(lockdown=service_provider).set_wifi_power_state(state == 'on')


@profile_group.command('erase-device', cls=Command)
@click.option('--preserve-data-plan/--no-preserve-data-plan', default=True,
              help='Preserves eSIM / data plan after erase')
@click.option('--disallow-proximity-setup/--no-disallow-proximity-setup', default=False,
              help='Disallows to setup the erased device from nearby devices')
def profile_erase_device(service_provider: LockdownServiceProvider, preserve_data_plan: bool,
                         disallow_proximity_setup: bool) -> None:
    """ Erase device """
    logger.info(f'Erasing device with preserve_data_plan: {preserve_data_plan}, '
                f'disallow_proximity_setup: {disallow_proximity_setup}')
    MobileConfigService(lockdown=service_provider).erase_device(preserve_data_plan, disallow_proximity_setup)
    logger.info('Erased device')


@profile_group.command('create-keybag')
@click.argument('keybag', type=click.Path(file_okay=True, dir_okay=False, exists=False))
@click.argument('organization')
def profile_create_keybag(keybag: str, organization: str) -> None:
    """ Create keybag storing certificate and private key """
    create_keybag_file(Path(keybag), organization)


@profile_group.command('supervise', cls=Command)
@click.argument('organization')
@click.option('--keybag', type=click.Path(file_okay=True, dir_okay=False, exists=True))
def profile_supervise(service_provider: LockdownServiceProvider, organization: str, keybag: Optional[str]) -> None:
    """ Supervise device """
    if MobileActivationService(service_provider).state == 'Unactivated':
        logger.info('Activating device')
        MobileActivationService(service_provider).activate()
        logger.info('Device has been successfully activated')
    logger.info('Supervising device')
    if keybag is None:
        with tempfile.TemporaryDirectory() as temp_dir:
            keybag = Path(temp_dir) / 'keybag'
            create_keybag_file(keybag, organization)
            MobileConfigService(lockdown=service_provider).supervise(organization, keybag)
    else:
        MobileConfigService(lockdown=service_provider).supervise(organization, Path(keybag))

    logger.info('Device has been successfully supervised')


@profile_group.command('install-wifi-profile', cls=Command)
@click.argument('encryption_type')
@click.argument('ssid')
@click.argument('password')
@click.option('--keybag', type=click.Path(file_okay=True, dir_okay=False, exists=True))
def profile_install_wifi_profile(service_provider: LockdownServiceProvider, encryption_type: str, ssid: str,
                                 password: str,
                                 keybag: Optional[str]) -> None:
    """
    Install Wi-Fi profile

    This will enable the device to auto-connect to given network
    """
    if keybag is not None:
        keybag = Path(keybag)
    MobileConfigService(lockdown=service_provider).install_wifi_profile(
        encryption_type=encryption_type, ssid=ssid, password=password, keybag_file=keybag)


@profile_group.command('install-http-proxy', cls=Command)
@click.argument('server')
@click.argument('port', type=click.IntRange(1, 65535))
@click.option('--keybag', type=click.Path(file_okay=True, dir_okay=False, exists=True))
def profile_install_http_proxy(service_provider: LockdownServiceProvider, server: str, port: int,
                               keybag: Optional[str]) -> None:
    """ Install HTTP Proxy profile """
    if keybag is not None:
        keybag = Path(keybag)
    MobileConfigService(lockdown=service_provider).install_http_proxy(server, port, keybag_file=keybag)


@profile_group.command('remove-http-proxy', cls=Command)
def profile_remove_http_proxy(service_provider: LockdownServiceProvider) -> None:
    """ Remove HTTP Proxy profile that was previously installed using pymobiledevice3 """
    MobileConfigService(lockdown=service_provider).remove_http_proxy()


@profile_group.command('install-restrictions-profile', cls=Command)
@click.option('--keybag', type=click.Path(file_okay=True, dir_okay=False, exists=True))
@click.option('--enforced-software-update-delay', type=click.IntRange(0, 90), default=0)
def profile_install_restrictions_profile(
        service_provider: LockdownServiceProvider, keybag: Optional[str], enforced_software_update_delay: int) -> None:
    """ Install restrictions profile (can be used for delayed OTA) """
    if keybag is not None:
        keybag = Path(keybag)
    MobileConfigService(lockdown=service_provider).install_restrictions_profile(
        enforced_software_update_delay=enforced_software_update_delay, keybag_file=keybag)



import inspect
import socket
import sys
from datetime import datetime
from pathlib import Path
from typing import List, Tuple

from pymobiledevice3.exceptions import FeatureNotSupportedError, OSNotSupportedError

DEFAULT_AFTER_IDLE_SEC = 3
DEFAULT_INTERVAL_SEC = 3
DEFAULT_MAX_FAILS = 3


class OsUtils:
    _instance = None
    _os_name = None

    @classmethod
    def create(cls) -> 'OsUtils':
        if cls._instance is None:
            cls._os_name = sys.platform
            if cls._os_name == 'win32':
                from pymobiledevice3.osu.win_util import Win32
                cls._instance = Win32()
            elif cls._os_name == 'darwin':
                from pymobiledevice3.osu.posix_util import Darwin
                cls._instance = Darwin()
            elif cls._os_name == 'linux':
                from pymobiledevice3.osu.posix_util import Linux
                cls._instance = Linux()
            elif cls._os_name == 'cygwin':
                from pymobiledevice3.osu.posix_util import Cygwin
                cls._instance = Cygwin()
            else:
                raise OSNotSupportedError(cls._os_name)
        return cls._instance

    @property
    def is_admin(self) -> bool:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def usbmux_address(self) -> Tuple[str, int]:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def bonjour_timeout(self) -> int:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def loopback_header(self) -> bytes:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def access_denied_error(self) -> str:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def pair_record_path(self) -> Path:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def get_ipv6_ips(self) -> List[str]:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def set_keepalive(self, sock: socket.socket, after_idle_sec: int = DEFAULT_AFTER_IDLE_SEC,
                      interval_sec: int = DEFAULT_INTERVAL_SEC, max_fails: int = DEFAULT_MAX_FAILS) -> None:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def parse_timestamp(self, time_stamp) -> datetime:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def chown_to_non_sudo_if_needed(self, path: Path) -> None:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def wait_return(self):
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def get_homedir(self) -> Path:
        return Path.home()


def get_os_utils() -> OsUtils:
    return OsUtils.create()



from pymobiledevice3.services.webinspector import SAFARI, WebinspectorService

TIMEOUT = 10


def test_opening_app(lockdown):
    inspector = WebinspectorService(lockdown=lockdown)
    inspector.connect(timeout=TIMEOUT)
    safari = inspector.open_app(SAFARI)
    pages = inspector.get_open_pages()
    # Might take a while to update.
    if safari.name not in pages:
        inspector.flush_input(1)
    pages = inspector.get_open_pages()
    try:
        assert safari.name in pages
        assert pages[safari.name]
    finally:
        inspector.close()
        inspector.loop.close()



import dataclasses
import uuid
from datetime import datetime
from typing import Any, List, Mapping

from construct import Aligned, Array, Bytes, Const, CString, Default, Double, Enum, ExprAdapter, FlagsEnum, \
    GreedyBytes, Hex, If, Int32ul, Int64sl, Int64ul, LazyBound
from construct import Optional as ConstructOptional
from construct import Pass, Prefixed, Probe, Struct, Switch, this

XpcMessageType = Enum(Hex(Int32ul),
                      NULL=0x00001000,
                      BOOL=0x00002000,
                      INT64=0x00003000,
                      UINT64=0x00004000,
                      DOUBLE=0x00005000,
                      POINTER=0x00006000,
                      DATE=0x00007000,
                      DATA=0x00008000,
                      STRING=0x00009000,
                      UUID=0x0000a000,
                      FD=0x0000b000,
                      SHMEM=0x0000c000,
                      MACH_SEND=0x0000d000,
                      ARRAY=0x0000e000,
                      DICTIONARY=0x0000f000,
                      ERROR=0x00010000,
                      CONNECTION=0x00011000,
                      ENDPOINT=0x00012000,
                      SERIALIZER=0x00013000,
                      PIPE=0x00014000,
                      MACH_RECV=0x00015000,
                      BUNDLE=0x00016000,
                      SERVICE=0x00017000,
                      SERVICE_INSTANCE=0x00018000,
                      ACTIVITY=0x00019000,
                      FILE_TRANSFER=0x0001a000,
                      )
XpcFlags = FlagsEnum(Hex(Int32ul),
                     ALWAYS_SET=0x00000001,
                     PING=0x00000002,
                     DATA_PRESENT=0x00000100,
                     WANTING_REPLY=0x00010000,
                     REPLY=0x00020000,
                     FILE_TX_STREAM_REQUEST=0x00100000,
                     FILE_TX_STREAM_RESPONSE=0x00200000,
                     INIT_HANDSHAKE=0x00400000,
                     )
AlignedString = Aligned(4, CString('utf8'))
XpcNull = Pass
XpcBool = Int32ul
XpcInt64 = Int64sl
XpcUInt64 = Int64ul
XpcDouble = Double
XpcPointer = None
XpcDate = Int64ul
XpcData = Aligned(4, Prefixed(Int32ul, GreedyBytes))
XpcString = Aligned(4, Prefixed(Int32ul, CString('utf8')))
XpcUuid = Bytes(16)
XpcFd = Int32ul
XpcShmem = Struct('length' / Int32ul, Int32ul)
XpcArray = Prefixed(Int32ul, Struct(
    'count' / Int32ul,
    'entries' / Array(this.count, LazyBound(lambda: XpcObject))))
XpcDictionaryEntry = Struct(
    'key' / AlignedString,
    'value' / LazyBound(lambda: XpcObject),
)
XpcDictionary = Prefixed(Int32ul, Struct(
    'count' / Hex(Int32ul),
    'entries' / If(this.count > 0, Array(this.count, XpcDictionaryEntry)),
))
XpcFileTransfer = Struct(
    'msg_id' / Int64ul,
    'data' / LazyBound(lambda: XpcObject),
)
XpcObject = Struct(
    'type' / XpcMessageType,
    'data' / Switch(this.type, {
        XpcMessageType.DICTIONARY: XpcDictionary,
        XpcMessageType.STRING: XpcString,
        XpcMessageType.INT64: XpcInt64,
        XpcMessageType.UINT64: XpcUInt64,
        XpcMessageType.DOUBLE: XpcDouble,
        XpcMessageType.BOOL: XpcBool,
        XpcMessageType.NULL: XpcNull,
        XpcMessageType.UUID: XpcUuid,
        XpcMessageType.POINTER: XpcPointer,
        XpcMessageType.DATE: XpcDate,
        XpcMessageType.DATA: XpcData,
        XpcMessageType.FD: XpcFd,
        XpcMessageType.SHMEM: XpcShmem,
        XpcMessageType.ARRAY: XpcArray,
        XpcMessageType.FILE_TRANSFER: XpcFileTransfer,
    }, default=Probe(lookahead=1000)),
)
XpcPayload = Struct(
    'magic' / Hex(Const(0x42133742, Int32ul)),
    'protocol_version' / Hex(Const(0x00000005, Int32ul)),
    'obj' / XpcObject,
)
XpcWrapper = Struct(
    'magic' / Hex(Const(0x29b00b92, Int32ul)),
    'flags' / Default(XpcFlags, XpcFlags.ALWAYS_SET),
    'message' / Prefixed(
        ExprAdapter(Int64ul, lambda obj, context: obj + 8, lambda obj, context: obj - 8),
        Struct(
            'message_id' / Hex(Default(Int64ul, 0)),
            'payload' / ConstructOptional(XpcPayload),
        ))
)


class XpcInt64Type(int):
    pass


class XpcUInt64Type(int):
    pass


@dataclasses.dataclass
class FileTransferType:
    transfer_size: int


def _decode_xpc_dictionary(xpc_object) -> Mapping:
    if xpc_object.data.count == 0:
        return {}
    result = {}
    for entry in xpc_object.data.entries:
        result[entry.key] = decode_xpc_object(entry.value)
    return result


def _decode_xpc_array(xpc_object) -> List:
    result = []
    for entry in xpc_object.data.entries:
        result.append(decode_xpc_object(entry))
    return result


def _decode_xpc_bool(xpc_object) -> bool:
    return bool(xpc_object.data)


def _decode_xpc_int64(xpc_object) -> XpcInt64Type:
    return XpcInt64Type(xpc_object.data)


def _decode_xpc_uint64(xpc_object) -> XpcUInt64Type:
    return XpcUInt64Type(xpc_object.data)


def _decode_xpc_uuid(xpc_object) -> uuid.UUID:
    return uuid.UUID(bytes=xpc_object.data)


def _decode_xpc_string(xpc_object) -> str:
    return xpc_object.data


def _decode_xpc_data(xpc_object) -> bytes:
    return xpc_object.data


def _decode_xpc_date(xpc_object) -> datetime:
    # Convert from nanoseconds to seconds
    return datetime.fromtimestamp(xpc_object.data / 1000000000)


def _decode_xpc_file_transfer(xpc_object) -> FileTransferType:
    return FileTransferType(transfer_size=_decode_xpc_dictionary(xpc_object.data.data)['s'])


def _decode_xpc_double(xpc_object) -> float:
    return xpc_object.data


def _decode_xpc_null(xpc_object) -> None:
    return None


def decode_xpc_object(xpc_object) -> Any:
    decoders = {
        XpcMessageType.DICTIONARY: _decode_xpc_dictionary,
        XpcMessageType.ARRAY: _decode_xpc_array,
        XpcMessageType.BOOL: _decode_xpc_bool,
        XpcMessageType.INT64: _decode_xpc_int64,
        XpcMessageType.UINT64: _decode_xpc_uint64,
        XpcMessageType.UUID: _decode_xpc_uuid,
        XpcMessageType.STRING: _decode_xpc_string,
        XpcMessageType.DATA: _decode_xpc_data,
        XpcMessageType.DATE: _decode_xpc_date,
        XpcMessageType.FILE_TRANSFER: _decode_xpc_file_transfer,
        XpcMessageType.DOUBLE: _decode_xpc_double,
        XpcMessageType.NULL: _decode_xpc_null,
    }
    decoder = decoders.get(xpc_object.type)
    if decoder is None:
        raise TypeError(f'deserialize error: {xpc_object}')
    return decoder(xpc_object)


def _build_xpc_array(payload: List) -> Mapping:
    entries = []
    for entry in payload:
        entry = _build_xpc_object(entry)
        entries.append(entry)
    return {
        'type': XpcMessageType.ARRAY,
        'data': {
            'count': len(entries),
            'entries': entries
        }
    }


def _build_xpc_dictionary(payload: Mapping) -> Mapping:
    entries = []
    for key, value in payload.items():
        entry = {'key': key, 'value': _build_xpc_object(value)}
        entries.append(entry)
    return {
        'type': XpcMessageType.DICTIONARY,
        'data': {
            'count': len(entries),
            'entries': entries,
        }
    }


def _build_xpc_bool(payload: bool) -> Mapping:
    return {
        'type': XpcMessageType.BOOL,
        'data': payload,
    }


def _build_xpc_string(payload: str) -> Mapping:
    return {
        'type': XpcMessageType.STRING,
        'data': payload,
    }


def _build_xpc_data(payload: bool) -> Mapping:
    return {
        'type': XpcMessageType.DATA,
        'data': payload,
    }


def _build_xpc_double(payload: float) -> Mapping:
    return {
        'type': XpcMessageType.DOUBLE,
        'data': payload,
    }


def _build_xpc_uuid(payload: uuid.UUID) -> Mapping:
    return {
        'type': XpcMessageType.UUID,
        'data': payload.bytes,
    }


def _build_xpc_null(payload: None) -> Mapping:
    return {
        'type': XpcMessageType.NULL,
        'data': None,
    }


def _build_xpc_uint64(payload: XpcUInt64Type) -> Mapping:
    return {
        'type': XpcMessageType.UINT64,
        'data': payload,
    }


def _build_xpc_int64(payload: XpcInt64Type) -> Mapping:
    return {
        'type': XpcMessageType.INT64,
        'data': payload,
    }


def _build_xpc_object(payload: Any) -> Mapping:
    if payload is None:
        return _build_xpc_null(payload)
    payload_builders = {
        list: _build_xpc_array,
        dict: _build_xpc_dictionary,
        bool: _build_xpc_bool,
        str: _build_xpc_string,
        bytes: _build_xpc_data,
        bytearray: _build_xpc_data,
        float: _build_xpc_double,
        uuid.UUID: _build_xpc_uuid,
        'XpcUInt64Type': _build_xpc_uint64,
        'XpcInt64Type': _build_xpc_int64,
    }
    builder = payload_builders.get(type(payload), payload_builders.get(type(payload).__name__))
    if builder is None:
        raise TypeError(f'unrecognized type for: {payload} {type(payload)}')
    return builder(payload)


def create_xpc_wrapper(d: Mapping, message_id: int = 0, wanting_reply: bool = False) -> bytes:
    flags = XpcFlags.ALWAYS_SET
    if len(d.keys()) > 0:
        flags |= XpcFlags.DATA_PRESENT
    if wanting_reply:
        flags |= XpcFlags.WANTING_REPLY

    xpc_payload = {
        'message_id': message_id,
        'payload': {'obj': _build_xpc_object(d)}
    }

    xpc_wrapper = {
        'flags': flags,
        'message': xpc_payload
    }
    return XpcWrapper.build(xpc_wrapper)



#!/usr/bin/env python3
import dataclasses
import plistlib
import struct
import tempfile
import typing
from datetime import datetime
from enum import IntEnum
from pathlib import Path
from tarfile import TarFile

from construct import Adapter, Byte, Bytes, Computed, Enum, Int16ul, Int32ul, Optional, RepeatUntil, Struct, this

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService
from pymobiledevice3.utils import try_decode

CHUNK_SIZE = 4096
TIME_FORMAT = '%H:%M:%S'
SYSLOG_LINE_SPLITTER = '\n\x00'


class SyslogLogLevel(IntEnum):
    NOTICE = 0x00
    INFO = 0x01
    DEBUG = 0x02

    # deducted from console-app
    USER_ACTION = 0x03
    ERROR = 0x10
    FAULT = 0x11


@dataclasses.dataclass
class SyslogLabel:
    category: str
    subsystem: str


@dataclasses.dataclass
class SyslogEntry:
    pid: int
    timestamp: datetime
    level: SyslogLogLevel
    image_name: str
    filename: str
    message: str
    label: typing.Optional[SyslogLabel] = None


class TimestampAdapter(Adapter):
    def _decode(self, obj, context, path):
        return datetime.fromtimestamp(obj.seconds + (obj.microseconds / 1000000))

    def _encode(self, obj, context, path):
        return list(map(int, obj.split(".")))


timestamp_t = Struct(
    'seconds' / Int32ul,
    Bytes(4),
    'microseconds' / Int32ul
)

syslog_t = Struct(
    Bytes(9),
    'pid' / Int32ul,
    Bytes(42),
    'timestamp' / TimestampAdapter(timestamp_t),
    Bytes(1),
    'level' / Enum(Byte, Notice=0, Info=0x01, Debug=0x02, Error=0x10, Fault=0x11),
    Bytes(38),
    'image_name_size' / Int16ul,
    'message_size' / Int16ul,
    Bytes(6),
    '_subsystem_size' / Int32ul,
    '_category_size' / Int32ul,
    Bytes(4),
    '_filename' / RepeatUntil(lambda x, lst, ctx: lst[-1] == 0, Byte),
    'filename' / Computed(lambda ctx: try_decode(bytearray(ctx._filename[:-1]))),
    '_image_name' / Bytes(this.image_name_size),
    'image_name' / Computed(lambda ctx: try_decode(ctx._image_name[:-1])),
    '_message' / Bytes(this.message_size),
    'message' / Computed(lambda ctx: try_decode(ctx._message[:-1])),
    'label' / Optional(Struct(
        '_subsystem' / Bytes(this._._subsystem_size),
        'subsystem' / Computed(lambda ctx: try_decode(ctx._subsystem[:-1])),
        '_category' / Bytes(this._._category_size),
        'category' / Computed(lambda ctx: try_decode(ctx._category[:-1])),
    )),
)


class OsTraceService(LockdownService):
    """
    Provides API for the following operations:
    * Show process list (process name and pid)
    * Stream syslog lines in binary form with optional filtering by pid.
    * Get old stored syslog archive in PAX format (can be extracted using `pax -r < filename`).
        * Archive contain the contents are the `/var/db/diagnostics` directory
    """
    SERVICE_NAME = 'com.apple.os_trace_relay'
    RSD_SERVICE_NAME = 'com.apple.os_trace_relay.shim.remote'

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    def get_pid_list(self):
        self.service.send_plist({'Request': 'PidList'})

        # ignore first received unknown byte
        self.service.recvall(1)

        response = self.service.recv_prefixed()
        return plistlib.loads(response)

    def create_archive(self, out: typing.IO, size_limit: int = None, age_limit: int = None, start_time: int = None):
        request = {'Request': 'CreateArchive'}

        if size_limit is not None:
            request.update({'SizeLimit': size_limit})

        if age_limit is not None:
            request.update({'AgeLimit': age_limit})

        if start_time is not None:
            request.update({'StartTime': start_time})

        self.service.send_plist(request)

        assert 1 == self.service.recvall(1)[0]

        assert plistlib.loads(self.service.recv_prefixed()).get('Status') == 'RequestSuccessful', 'Invalid status'

        while True:
            try:
                assert 3 == self.service.recvall(1)[0], 'invalid magic'
            except ConnectionAbortedError:
                break
            out.write(self.service.recv_prefixed(endianity='<'))

    def collect(self, out: str, size_limit: typing.Optional[int] = None, age_limit: typing.Optional[int] = None,
                start_time: typing.Optional[int] = None) -> None:
        """
        Collect the system logs into a .logarchive that can be viewed later with tools such as log or Console.

        :param out: output file name
        :param size_limit: maximum size in bytes of logarchive
        :param age_limit: maximum age in days
        :param start_time: start time of logarchive in unix timestamp
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            file = Path(temp_dir) / 'foo.tar'
            with open(file, 'wb') as f:
                self.create_archive(f, size_limit=size_limit, age_limit=age_limit, start_time=start_time)
            TarFile(file).extractall(out)

    def syslog(self, pid=-1) -> typing.Generator[SyslogEntry, None, None]:
        self.service.send_plist({'Request': 'StartActivity', 'MessageFilter': 65535, 'Pid': pid, 'StreamFlags': 60})

        length_length, = struct.unpack('<I', self.service.recvall(4))
        length = int(self.service.recvall(length_length)[::-1].hex(), 16)
        response = plistlib.loads(self.service.recvall(length))

        if response.get('Status') != 'RequestSuccessful':
            raise PyMobileDevice3Exception(f'got invalid response: {response}')

        while True:
            assert b'\x02' == self.service.recvall(1)
            length, = struct.unpack('<I', self.service.recvall(4))
            line = self.service.recvall(length)
            entry = syslog_t.parse(line)
            label = None
            if entry.label is not None:
                label = SyslogLabel(subsystem=entry.label.subsystem, category=entry.label.category)
            yield SyslogEntry(pid=entry.pid, timestamp=entry.timestamp, level=SyslogLogLevel(int(entry.level)),
                              image_name=entry.image_name, filename=entry.filename, message=entry.message, label=label)



import logging
import sys
import traceback

import click
import coloredlogs

from pymobiledevice3.exceptions import AccessDeniedError, CloudConfigurationAlreadyPresentError, \
    ConnectionFailedToUsbmuxdError, DeprecationError, DeveloperModeError, DeveloperModeIsNotEnabledError, \
    DeviceHasPasscodeSetError, DeviceNotFoundError, FeatureNotSupportedError, InternalError, InvalidServiceError, \
    MessageNotSupportedError, MissingValueError, NoDeviceConnectedError, NotEnoughDiskSpaceError, NotPairedError, \
    OSNotSupportedError, PairingDialogResponsePendingError, PasswordRequiredError, RSDRequiredError, \
    SetProhibitedError, TunneldConnectionError, UserDeniedPairingError
from pymobiledevice3.osu.os_utils import get_os_utils

coloredlogs.install(level=logging.INFO)

logging.getLogger('quic').disabled = True
logging.getLogger('asyncio').disabled = True
logging.getLogger('zeroconf').disabled = True
logging.getLogger('parso.cache').disabled = True
logging.getLogger('parso.cache.pickle').disabled = True
logging.getLogger('parso.python.diff').disabled = True
logging.getLogger('humanfriendly.prompts').disabled = True
logging.getLogger('blib2to3.pgen2.driver').disabled = True
logging.getLogger('urllib3.connectionpool').disabled = True

logger = logging.getLogger(__name__)

INVALID_SERVICE_MESSAGE = """Failed to start service. Possible reasons are:
- If you were trying to access a developer service (developer subcommand):
    - If your device iOS version >= 15.0:
        - Make sure you first enabled "Developer Mode" via:
          > python3 -m pymobiledevice3 amfi enable-developer-mode

    - Make sure the DeveloperDiskImage/PersonalizedImage is mounted via:
      > python3 -m pymobiledevice3 mounter auto-mount

    - If your device iOS version >= 17.0:
        - Make sure you passed the --rsd option to the subcommand
          https://github.com/doronz88/pymobiledevice3#working-with-developer-tools-ios--170

- Apple removed this service

- A bug. Please file a bug report:
  https://github.com/doronz88/pymobiledevice3/issues/new?assignees=&labels=&projects=&template=bug_report.md&title=
"""

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'], max_content_width=400)

# Mapping of index options to import file names
CLI_GROUPS = {
    'activation': 'activation',
    'afc': 'afc',
    'amfi': 'amfi',
    'apps': 'apps',
    'backup2': 'backup',
    'bonjour': 'bonjour',
    'companion': 'companion_proxy',
    'crash': 'crash',
    'developer': 'developer',
    'diagnostics': 'diagnostics',
    'lockdown': 'lockdown',
    'mounter': 'mounter',
    'notification': 'notification',
    'pcap': 'pcap',
    'power-assertion': 'power_assertion',
    'processes': 'processes',
    'profile': 'profile',
    'provision': 'provision',
    'remote': 'remote',
    'restore': 'restore',
    'springboard': 'springboard',
    'syslog': 'syslog',
    'usbmux': 'usbmux',
    'webinspector': 'webinspector',
    'version': 'version',
}


class Pmd3Cli(click.Group):
    def list_commands(self, ctx):
        return CLI_GROUPS.keys()

    def get_command(self, ctx, name):
        if name not in CLI_GROUPS.keys():
            ctx.fail(f'No such command {name!r}.')
        mod = __import__(f'pymobiledevice3.cli.{CLI_GROUPS[name]}', None, None, ['cli'])
        command = mod.cli.get_command(ctx, name)
        # Some cli groups have different names than the index
        if not command:
            command_name = mod.cli.list_commands(ctx)[0]
            command = mod.cli.get_command(ctx, command_name)
        return command


@click.command(cls=Pmd3Cli, context_settings=CONTEXT_SETTINGS)
def cli():
    """
    \b
    Interact with a connected iDevice (iPhone, iPad, ...)
    For more information please look at:
        https://github.com/doronz88/pymobiledevice3
    """
    pass


def main() -> None:
    try:
        cli()
    except NoDeviceConnectedError:
        logger.error('Device is not connected')
    except ConnectionAbortedError:
        logger.error('Device was disconnected')
    except NotPairedError:
        logger.error('Device is not paired')
    except UserDeniedPairingError:
        logger.error('User refused to trust this computer')
    except PairingDialogResponsePendingError:
        logger.error('Waiting for user dialog approval')
    except SetProhibitedError:
        logger.error('lockdownd denied the access')
    except MissingValueError:
        logger.error('No such value')
    except DeviceHasPasscodeSetError:
        logger.error('Cannot enable developer-mode when passcode is set')
    except DeveloperModeError as e:
        logger.error(f'Failed to enable developer-mode. Error: {e}')
    except ConnectionFailedToUsbmuxdError:
        logger.error('Failed to connect to usbmuxd socket. Make sure it\'s running.')
    except MessageNotSupportedError:
        logger.error('Message not supported for this iOS version')
        traceback.print_exc()
    except InternalError:
        logger.error('Internal Error')
    except DeveloperModeIsNotEnabledError:
        logger.error('Developer Mode is disabled. You can try to enable it using: '
                     'python3 -m pymobiledevice3 amfi enable-developer-mode')
    except (InvalidServiceError, RSDRequiredError) as e:
        should_retry_over_tunneld = False
        if isinstance(e, RSDRequiredError):
            logger.warning('Trying again over tunneld since RSD is required for this command')
            should_retry_over_tunneld = True
        elif (e.identifier is not None) and ('developer' in sys.argv) and ('--tunnel' not in sys.argv):
            logger.warning('Got an InvalidServiceError. Trying again over tunneld since it is a developer command')
            should_retry_over_tunneld = True
        if should_retry_over_tunneld:
            if '--' in sys.argv:
                escape_sequence = sys.argv.index('--')
                before = sys.argv[:escape_sequence]
                after = sys.argv[escape_sequence:]
                sys.argv = before + ['--tunnel', e.identifier] + after
            else:
                sys.argv += ['--tunnel', e.identifier]
            return main()
        logger.error(INVALID_SERVICE_MESSAGE)
    except PasswordRequiredError:
        logger.error('Device is password protected. Please unlock and retry')
    except AccessDeniedError:
        logger.error(get_os_utils().access_denied_error)
    except BrokenPipeError:
        traceback.print_exc()
    except TunneldConnectionError:
        logger.error(
            'Unable to connect to Tunneld. You can start one using:\n'
            'sudo python3 -m pymobiledevice3 remote tunneld')
    except DeviceNotFoundError as e:
        logger.error(f'Device not found: {e.udid}')
    except NotEnoughDiskSpaceError:
        logger.error('Not enough disk space')
    except DeprecationError:
        logger.error('failed to query MobileGestalt, MobileGestalt deprecated (iOS >= 17.4).')
    except OSNotSupportedError as e:
        logger.error(
            f'Unsupported OS - {e.os_name}. To add support, consider contributing at '
            f'https://github.com/doronz88/pymobiledevice3.')
    except CloudConfigurationAlreadyPresentError:
        logger.error('A cloud configuration is already present on device. You must first erase the device in order '
                     'to install new one:\n'
                     '> pymobiledevice3 profile erase-device')
    except FeatureNotSupportedError as e:
        logger.error(
            f'Missing implementation of `{e.feature}` on `{e.os_name}`. To add support, consider contributing at '
            f'https://github.com/doronz88/pymobiledevice3.')


if __name__ == '__main__':
    main()



import struct

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.dvt.instruments.location_simulation_base import LocationSimulationBase
from pymobiledevice3.services.lockdown_service import LockdownService


class DtSimulateLocation(LockdownService, LocationSimulationBase):
    SERVICE_NAME = 'com.apple.dt.simulatelocation'

    def __init__(self, lockdown: LockdownClient):
        LockdownService.__init__(self, lockdown, self.SERVICE_NAME)
        LocationSimulationBase.__init__(self)

    def clear(self) -> None:
        """ stop simulation """
        service = self.lockdown.start_lockdown_developer_service(self.SERVICE_NAME)
        service.sendall(struct.pack('>I', 1))

    def set(self, latitude: float, longitude: float) -> None:
        """ stop simulation """
        service = self.lockdown.start_lockdown_developer_service(self.SERVICE_NAME)
        service.sendall(struct.pack('>I', 0))
        latitude = str(latitude).encode()
        longitude = str(longitude).encode()
        service.sendall(struct.pack('>I', len(latitude)) + latitude)
        service.sendall(struct.pack('>I', len(longitude)) + longitude)



import asyncio

import pytest

from pymobiledevice3.remote.common import ConnectionType
from pymobiledevice3.remote.tunnel_service import get_core_device_tunnel_services, get_remote_pairing_tunnel_services, \
    start_tunnel


@pytest.mark.parametrize('connection_type', [ConnectionType.USB, ConnectionType.WIFI])
@pytest.mark.asyncio
async def test_start_tunnel(connection_type: ConnectionType):
    get_tunnel_services = {
        connection_type.USB: get_core_device_tunnel_services,
        connection_type.WIFI: get_remote_pairing_tunnel_services,
    }
    tunnel_services = await get_tunnel_services[connection_type]()
    async with start_tunnel(tunnel_services[0]):
        await asyncio.sleep(1)
    for tunnel_service in tunnel_services:
        await tunnel_service.close()



import uuid
from typing import Any, Mapping

from pymobiledevice3.exceptions import CoreDeviceError
from pymobiledevice3.remote.remote_service import RemoteService
from pymobiledevice3.remote.xpc_message import XpcInt64Type, XpcUInt64Type


def _generate_core_device_version_dict(version: str) -> Mapping:
    version_components = version.split('.')
    return {'components': [XpcUInt64Type(component) for component in version_components],
            'originalComponentsCount': XpcInt64Type(len(version_components)),
            'stringValue': version}


CORE_DEVICE_VERSION = _generate_core_device_version_dict('325.3')


class CoreDeviceService(RemoteService):
    async def invoke(self, feature_identifier: str, input_: Mapping = None) -> Any:
        if input_ is None:
            input_ = {}
        response = await self.service.send_receive_request({
            'CoreDevice.CoreDeviceDDIProtocolVersion': XpcInt64Type(0),
            'CoreDevice.action': {},
            'CoreDevice.coreDeviceVersion': CORE_DEVICE_VERSION,
            'CoreDevice.deviceIdentifier': str(uuid.uuid4()),
            'CoreDevice.featureIdentifier': feature_identifier,
            'CoreDevice.input': input_,
            'CoreDevice.invocationIdentifier': str(uuid.uuid4())})
        output = response.get('CoreDevice.output')
        if output is None:
            raise CoreDeviceError(f'Failed to invoke: {feature_identifier}. Got error: {response}')
        return output



from pymobiledevice3.services.springboard import SpringBoardServicesService

PNG_HEADER = b'\x89\x50\x4e\x47\x0d\x0a\x1a\x0a'


def test_get_icon_png_data(lockdown):
    """
    Test that getting icon's data returns a valid PNG.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with SpringBoardServicesService(lockdown) as springboard:
        icon_data = springboard.get_icon_pngdata("com.apple.weather")
        assert icon_data.startswith(PNG_HEADER)


def test_get_icon_date(lockdown):
    """
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with SpringBoardServicesService(lockdown) as springboard:
        assert len(springboard.get_icon_state()[0]) > 0


def test_set_icon_date(lockdown):
    """
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    with SpringBoardServicesService(lockdown) as springboard:
        icon_state = springboard.get_icon_state()
        # swap docker icons
        icon_state[0] = icon_state[0][::-1]
        springboard.set_icon_state(icon_state)

        assert springboard.get_icon_state()[0] == icon_state[0]

        icon_state[0] = icon_state[0][::-1]
        springboard.set_icon_state(icon_state)

        assert springboard.get_icon_state()[0] == icon_state[0]



import logging
import os
import site
import sys
from pathlib import Path

import coloredlogs
import PyInstaller.__main__

import pymobiledevice3.__main__
import pymobiledevice3.cli
import pymobiledevice3.resources

coloredlogs.install(level=logging.DEBUG)

ROOT = Path(__file__).parent.parent.parent
DEFAULT_OUTPUT = ROOT / Path('dist/__main__').with_suffix('.exe' if sys.platform == 'win32' else '')
OUTPUT = (ROOT / 'dist/pymobiledevice3').with_suffix('.exe' if sys.platform == 'win32' else '')


def main() -> None:
    site_packages_path = site.getsitepackages()[0]
    resources_dir = Path(pymobiledevice3.resources.__file__).parent
    pymobiledevice3_cli_path = Path(pymobiledevice3.cli.__file__).parent

    hidden_imports = []

    for module in pymobiledevice3_cli_path.iterdir():
        if module.name.endswith('.py') and module.name != '__init__.py':  # Avoid including the __init__.py
            # Create the module name to be added to hidden imports
            module_name = 'pymobiledevice3.cli.' + os.path.splitext(module.name)[0]
            hidden_imports.append('--hidden-import=' + module_name)

    pyinstaller_args = [
        pymobiledevice3.__main__.__file__,
        '--hidden-import=ipsw_parser',
        '--hidden-import=zeroconf',
        '--hidden-import=pyimg4',
        '--hidden-import=zeroconf._utils.ipaddress',
        '--hidden-import=zeroconf._handlers.answers',
        '--hidden-import=readchar',
        '--copy-metadata=pyimg4',
        '--copy-metadata=readchar',
        '--copy-metadata=apple_compress',
        '--onefile',
    ]
    if sys.platform == 'win32':
        pyinstaller_args.extend([
            '--add-binary', f'{site_packages_path}/Lib/site-packages/pytun_pmd3/wintun/*;pytun_pmd3/wintun/bin',
            '--add-binary', f'{resources_dir}/webinspector;pymobiledevice3/resources/webinspector',
        ])
    else:
        pyinstaller_args.extend([
            '--add-binary', f'{site_packages_path}/pytun_pmd3:pytun_pmd3',
            '--add-binary', f'{resources_dir}/webinspector:pymobiledevice3/resources/webinspector',
        ])
    pyinstaller_args.extend(hidden_imports)
    PyInstaller.__main__.run(pyinstaller_args)
    DEFAULT_OUTPUT.rename(OUTPUT)


if __name__ == '__main__':
    main()



import asyncio
import dataclasses
import json
import logging
import os
import signal
import traceback
from contextlib import asynccontextmanager, suppress
from typing import Dict, List, Mapping, Optional, Tuple, Union

import construct
import fastapi
import requests
import uvicorn
from construct import StreamError
from fastapi import FastAPI
from packaging.version import Version

from pymobiledevice3 import usbmux
from pymobiledevice3.bonjour import REMOTED_SERVICE_NAMES, browse
from pymobiledevice3.exceptions import ConnectionFailedError, ConnectionFailedToUsbmuxdError, DeviceNotFoundError, \
    GetProhibitedError, InvalidServiceError, MuxException, PairingError, TunneldConnectionError
from pymobiledevice3.lockdown import create_using_usbmux, get_mobdev2_lockdowns
from pymobiledevice3.osu.os_utils import get_os_utils
from pymobiledevice3.remote.common import TunnelProtocol
from pymobiledevice3.remote.module_imports import start_tunnel
from pymobiledevice3.remote.remote_service_discovery import RSD_PORT, RemoteServiceDiscoveryService
from pymobiledevice3.remote.tunnel_service import CoreDeviceTunnelProxy, RemotePairingProtocol, TunnelResult, \
    create_core_device_tunnel_service_using_rsd, get_remote_pairing_tunnel_services
from pymobiledevice3.remote.utils import get_rsds, stop_remoted
from pymobiledevice3.utils import asyncio_print_traceback, get_asyncio_loop

logger = logging.getLogger(__name__)

TUNNELD_DEFAULT_ADDRESS = ('127.0.0.1', 49151)

# bugfix: after the device reboots, it might take some time for remoted to start answering the bonjour queries
REATTEMPT_INTERVAL = 5
REATTEMPT_COUNT = 5

REMOTEPAIRING_INTERVAL = 5
MOVDEV2_INTERVAL = 5

USBMUX_INTERVAL = 2
OSUTILS = get_os_utils()


@dataclasses.dataclass
class TunnelTask:
    task: asyncio.Task
    udid: Optional[str] = None
    tunnel: Optional[TunnelResult] = None


class TunneldCore:
    def __init__(self, protocol: TunnelProtocol = TunnelProtocol.QUIC, wifi_monitor: bool = True,
                 usb_monitor: bool = True, usbmux_monitor: bool = True, mobdev2_monitor: bool = True) -> None:
        self.protocol = protocol
        self.tasks: List[asyncio.Task] = []
        self.tunnel_tasks: Dict[str, TunnelTask] = {}
        self.usb_monitor = usb_monitor
        self.wifi_monitor = wifi_monitor
        self.usbmux_monitor = usbmux_monitor
        self.mobdev2_monitor = mobdev2_monitor

    def start(self) -> None:
        """ Register all tasks """
        self.tasks = []
        if self.usb_monitor:
            self.tasks.append(asyncio.create_task(self.monitor_usb_task(), name='monitor-usb-task'))
        if self.wifi_monitor:
            self.tasks.append(asyncio.create_task(self.monitor_wifi_task(), name='monitor-wifi-task'))
        if self.usbmux_monitor:
            self.tasks.append(asyncio.create_task(self.monitor_usbmux_task(), name='monitor-usbmux-task'))
        if self.mobdev2_monitor:
            self.tasks.append(asyncio.create_task(self.monitor_mobdev2_task(), name='monitor-mobdev2-task'))

    def tunnel_exists_for_udid(self, udid: str) -> bool:
        for task in self.tunnel_tasks.values():
            if (task.udid == udid) and (task.tunnel is not None):
                return True
        return False

    @asyncio_print_traceback
    async def monitor_usb_task(self) -> None:
        previous_ips = []
        while True:
            current_ips = OSUTILS.get_ipv6_ips()
            added = [ip for ip in current_ips if ip not in previous_ips]
            removed = [ip for ip in previous_ips if ip not in current_ips]

            previous_ips = current_ips

            logger.debug(f'added interfaces: {added}')
            logger.debug(f'removed interfaces: {removed}')

            for ip in removed:
                if ip in self.tunnel_tasks:
                    self.tunnel_tasks[ip].task.cancel()
                    await self.tunnel_tasks[ip].task

            for ip in added:
                self.tunnel_tasks[ip] = TunnelTask(
                    task=asyncio.create_task(self.handle_new_potential_usb_cdc_ncm_interface_task(ip),
                                             name=f'handle-new-potential-usb-cdc-ncm-interface-task-{ip}'))

            # wait before re-iterating
            await asyncio.sleep(1)

    @asyncio_print_traceback
    async def monitor_wifi_task(self) -> None:
        try:
            while True:
                for service in await get_remote_pairing_tunnel_services():
                    if service.hostname in self.tunnel_tasks:
                        # skip tunnel if already exists for this ip
                        await service.close()
                        continue
                    if self.tunnel_exists_for_udid(service.remote_identifier):
                        # skip tunnel if already exists for this udid
                        await service.close()
                        continue
                    self.tunnel_tasks[service.hostname] = TunnelTask(
                        task=asyncio.create_task(self.start_tunnel_task(service.hostname, service),
                                                 name=f'start-tunnel-task-wifi-{service.hostname}'),
                        udid=service.remote_identifier
                    )
                await asyncio.sleep(REMOTEPAIRING_INTERVAL)
        except asyncio.CancelledError:
            pass

    @asyncio_print_traceback
    async def monitor_usbmux_task(self) -> None:
        try:
            while True:
                try:
                    for mux_device in usbmux.list_devices():
                        task_identifier = f'usbmux-{mux_device.serial}-{mux_device.connection_type}'
                        if self.tunnel_exists_for_udid(mux_device.serial):
                            continue
                        try:
                            service = CoreDeviceTunnelProxy(create_using_usbmux(mux_device.serial))
                        except (MuxException, InvalidServiceError, GetProhibitedError, construct.core.StreamError,
                                ConnectionAbortedError, DeviceNotFoundError):
                            continue
                        self.tunnel_tasks[task_identifier] = TunnelTask(
                            udid=mux_device.serial,
                            task=asyncio.create_task(
                                self.start_tunnel_task(task_identifier,
                                                       service,
                                                       protocol=TunnelProtocol.TCP),
                                name=f'start-tunnel-task-{task_identifier}'))
                except ConnectionFailedToUsbmuxdError:
                    # This is exception is expected to occur repeatedly on linux running usbmuxd
                    # as long as there isn't any physical iDevice connected
                    logger.debug('failed to connect to usbmux. waiting for it to restart')
                finally:
                    await asyncio.sleep(USBMUX_INTERVAL)
        except asyncio.CancelledError:
            pass

    @asyncio_print_traceback
    async def monitor_mobdev2_task(self) -> None:
        try:
            while True:
                async for ip, lockdown in get_mobdev2_lockdowns(only_paired=True):
                    if self.tunnel_exists_for_udid(lockdown.udid):
                        # skip tunnel if already exists for this udid
                        continue
                    task_identifier = f'mobdev2-{lockdown.udid}-{ip}'
                    try:
                        tunnel_service = CoreDeviceTunnelProxy(lockdown)
                    except InvalidServiceError:
                        logger.warning(f'[{task_identifier}] failed to start CoreDeviceTunnelProxy - skipping')
                        continue
                    self.tunnel_tasks[task_identifier] = TunnelTask(
                        task=asyncio.create_task(self.start_tunnel_task(task_identifier, tunnel_service),
                                                 name=f'start-tunnel-task-{task_identifier}'),
                        udid=lockdown.udid
                    )
                await asyncio.sleep(MOVDEV2_INTERVAL)
        except asyncio.CancelledError:
            pass

    @asyncio_print_traceback
    async def start_tunnel_task(
            self, task_identifier: str, protocol_handler: Union[RemotePairingProtocol, CoreDeviceTunnelProxy],
            queue: Optional[asyncio.Queue] = None, protocol: Optional[TunnelProtocol] = None) -> None:
        if protocol is None:
            protocol = self.protocol
        if isinstance(protocol_handler, CoreDeviceTunnelProxy):
            protocol = TunnelProtocol.TCP
        tun = None
        bailed_out = False
        try:
            if self.tunnel_exists_for_udid(protocol_handler.remote_identifier):
                # cancel current tunnel creation
                raise asyncio.CancelledError()

            async with start_tunnel(protocol_handler, protocol=protocol) as tun:
                if not self.tunnel_exists_for_udid(protocol_handler.remote_identifier):
                    self.tunnel_tasks[task_identifier].tunnel = tun
                    self.tunnel_tasks[task_identifier].udid = protocol_handler.remote_identifier
                    if queue is not None:
                        queue.put_nowait(tun)
                        # avoid sending another message if succeeded
                        queue = None
                    logger.info(f'[{asyncio.current_task().get_name()}] Created tunnel --rsd {tun.address} {tun.port}')
                    await tun.client.wait_closed()
                else:
                    bailed_out = True
                    logger.debug(
                        f'not establishing tunnel from {asyncio.current_task().get_name()} '
                        f'since there is already an active one for same udid')
        except asyncio.CancelledError:
            pass
        except (ConnectionResetError, StreamError) as e:
            logger.debug(f'got {e.__class__.__name__} from {asyncio.current_task().get_name()}')
        except (asyncio.exceptions.IncompleteReadError, TimeoutError, OSError) as e:
            logger.debug(f'got {e.__class__.__name__} from tunnel --rsd {tun.address} {tun.port}')
        except Exception:
            logger.error(f'got exception from {asyncio.current_task().get_name()}: {traceback.format_exc()}')
        finally:
            if queue is not None:
                # notify something went wrong
                queue.put_nowait(None)

            if tun is not None and not bailed_out:
                logger.info(f'disconnected from tunnel --rsd {tun.address} {tun.port}')
                await tun.client.stop_tunnel()

            if protocol_handler is not None:
                try:
                    await protocol_handler.close()
                except OSError:
                    pass

            if task_identifier in self.tunnel_tasks:
                # in case the tunnel was removed just now
                self.tunnel_tasks.pop(task_identifier)

    @asyncio_print_traceback
    async def handle_new_potential_usb_cdc_ncm_interface_task(self, ip: str) -> None:
        rsd = None
        try:
            answers = None
            for i in range(REATTEMPT_COUNT):
                answers = await browse(REMOTED_SERVICE_NAMES, [ip])
                if answers:
                    break
                logger.debug(f'No addresses found for: {ip}')
                await asyncio.sleep(REATTEMPT_INTERVAL)

            if not answers:
                raise asyncio.CancelledError()

            peer_address = answers[0].ips[0]

            # establish an untrusted RSD handshake
            rsd = RemoteServiceDiscoveryService((peer_address, RSD_PORT))

            with stop_remoted():
                try:
                    await rsd.connect()
                except (ConnectionRefusedError, TimeoutError):
                    raise asyncio.CancelledError()

            if (self.protocol == TunnelProtocol.QUIC) and (Version(rsd.product_version) < Version('17.0.0')):
                await rsd.close()
                raise asyncio.CancelledError()

            await asyncio.create_task(
                self.start_tunnel_task(ip, await create_core_device_tunnel_service_using_rsd(rsd)),
                name=f'start-tunnel-task-usb-{ip}')
        except asyncio.CancelledError:
            pass
        except PairingError as e:
            logger.error(f'Failed to pair with {ip} with error: {e}')
        except RuntimeError:
            logger.debug(f'Got RuntimeError from: {asyncio.current_task().get_name()}')
        except Exception:
            logger.error(f'Error raised from: {asyncio.current_task().get_name()}: {traceback.format_exc()}')
        finally:
            if rsd is not None:
                try:
                    await rsd.close()
                except OSError:
                    pass

            if ip in self.tunnel_tasks:
                # in case the tunnel was removed just now
                self.tunnel_tasks.pop(ip)

    async def close(self) -> None:
        """ close all tasks """
        for task in self.tasks + [tunnel_task.task for tunnel_task in self.tunnel_tasks.values()]:
            task.cancel()
            with suppress(asyncio.CancelledError):
                await task

    def get_tunnels_ips(self) -> Dict:
        """ Retrieve the available tunnel tasks and format them as {UDID: [IP]} """
        tunnels_ips = {}
        for ip, active_tunnel in self.tunnel_tasks.items():
            if (active_tunnel.udid is None) or (active_tunnel.tunnel is None):
                continue
            if active_tunnel.udid not in tunnels_ips:
                tunnels_ips[active_tunnel.udid] = [ip]
            else:
                tunnels_ips[active_tunnel.udid].append(ip)
        return tunnels_ips

    def cancel(self, udid: str) -> None:
        """ Cancel active tunnels """
        for tunnel_ip in self.get_tunnels_ips().get(udid, []):
            self.tunnel_tasks.pop(tunnel_ip).task.cancel()
            logger.info(f'canceling tunnel {tunnel_ip}')

    def clear(self) -> None:
        """ Clear active tunnels """
        for udid, tunnel in self.tunnel_tasks.items():
            logger.info(f'Removing tunnel {tunnel}')
            tunnel.task.cancel()
        self.tunnel_tasks = {}


class TunneldRunner:
    """ TunneldRunner orchestrate between the webserver and TunneldCore """

    @classmethod
    def create(cls, host: str, port: int, protocol: TunnelProtocol = TunnelProtocol.QUIC, usb_monitor: bool = True,
               wifi_monitor: bool = True, usbmux_monitor: bool = True, mobdev2_monitor: bool = True) -> None:
        cls(host, port, protocol=protocol, usb_monitor=usb_monitor, wifi_monitor=wifi_monitor,
            usbmux_monitor=usbmux_monitor, mobdev2_monitor=mobdev2_monitor)._run_app()

    def __init__(self, host: str, port: int, protocol: TunnelProtocol = TunnelProtocol.QUIC, usb_monitor: bool = True,
                 wifi_monitor: bool = True, usbmux_monitor: bool = True, mobdev2_monitor: bool = True):
        @asynccontextmanager
        async def lifespan(app: FastAPI):
            logging.getLogger('zeroconf').disabled = True
            self._tunneld_core.start()
            yield
            logger.info('Closing tunneld tasks...')
            await self._tunneld_core.close()

        self.host = host
        self.port = port
        self.protocol = protocol
        self._app = FastAPI(lifespan=lifespan)
        self._tunneld_core = TunneldCore(protocol=protocol, wifi_monitor=wifi_monitor, usb_monitor=usb_monitor,
                                         usbmux_monitor=usbmux_monitor, mobdev2_monitor=mobdev2_monitor)

        @self._app.get('/')
        async def list_tunnels() -> Mapping[str, List[Mapping]]:
            """ Retrieve the available tunnels and format them as {UUID: TUNNEL_ADDRESS} """
            tunnels = {}
            for ip, active_tunnel in self._tunneld_core.tunnel_tasks.items():
                if (active_tunnel.udid is None) or (active_tunnel.tunnel is None):
                    continue
                if active_tunnel.udid not in tunnels:
                    tunnels[active_tunnel.udid] = []
                tunnels[active_tunnel.udid].append(({
                    'tunnel-address': active_tunnel.tunnel.address,
                    'tunnel-port': active_tunnel.tunnel.port,
                    'interface': ip}))
            return tunnels

        @self._app.get('/shutdown')
        async def shutdown() -> fastapi.Response:
            """ Shutdown Tunneld """
            os.kill(os.getpid(), signal.SIGINT)
            data = {'operation': 'shutdown', 'data': True, 'message': 'Server shutting down...'}
            return generate_http_response(data)

        @self._app.get('/clear_tunnels')
        async def clear_tunnels() -> fastapi.Response:
            self._tunneld_core.clear()
            data = {'operation': 'clear_tunnels', 'data': True, 'message': 'Cleared tunnels...'}
            return generate_http_response(data)

        @self._app.get('/cancel')
        async def cancel_tunnel(udid: str) -> fastapi.Response:
            self._tunneld_core.cancel(udid=udid)
            data = {'operation': 'cancel', 'udid': udid, 'data': True, 'message': f'tunnel {udid} Canceled ...'}
            return generate_http_response(data)

        @self._app.get('/hello')
        async def hello() -> fastapi.Response:
            data = {'message': 'Hello, I\'m alive'}
            return generate_http_response(data)

        def generate_http_response(
                data: dict, status_code: int = 200, media_type: str = "application/json") -> fastapi.Response:
            return fastapi.Response(
                status_code=status_code,
                media_type=media_type,
                content=json.dumps(data))

        @self._app.get('/start-tunnel')
        async def start_tunnel(
                udid: str, ip: Optional[str] = None, connection_type: Optional[str] = None) -> fastapi.Response:
            udid_tunnels = [t.tunnel for t in self._tunneld_core.tunnel_tasks.values() if t.udid == udid]
            if len(udid_tunnels) > 0:
                data = {
                    'interface': udid_tunnels[0].interface,
                    'port': udid_tunnels[0].port,
                    'address': udid_tunnels[0].address
                }
                return generate_http_response(data)

            queue = asyncio.Queue()
            created_task = False

            try:
                if not created_task and connection_type in ('usbmux', None):
                    task_identifier = f'usbmux-{udid}'
                    try:
                        service = CoreDeviceTunnelProxy(create_using_usbmux(udid))
                        task = asyncio.create_task(
                            self._tunneld_core.start_tunnel_task(task_identifier, service, protocol=TunnelProtocol.TCP,
                                                                 queue=queue),
                            name=f'start-tunnel-task-{task_identifier}')
                        self._tunneld_core.tunnel_tasks[task_identifier] = TunnelTask(task=task, udid=udid)
                        created_task = True
                    except (ConnectionFailedError, InvalidServiceError, MuxException):
                        pass
                if connection_type in ('usb', None):
                    for rsd in await get_rsds(udid=udid):
                        rsd_ip = rsd.service.address[0]
                        if ip is not None and rsd_ip != ip:
                            await rsd.close()
                            continue
                        task = asyncio.create_task(
                            self._tunneld_core.start_tunnel_task(rsd_ip,
                                                                 await create_core_device_tunnel_service_using_rsd(rsd),
                                                                 queue=queue),
                            name=f'start-tunnel-usb-{rsd_ip}')
                        self._tunneld_core.tunnel_tasks[rsd_ip] = TunnelTask(task=task, udid=rsd.udid)
                        created_task = True
                if not created_task and connection_type in ('wifi', None):
                    for remotepairing in await get_remote_pairing_tunnel_services(udid=udid):
                        remotepairing_ip = remotepairing.hostname
                        if ip is not None and remotepairing_ip != ip:
                            await remotepairing.close()
                            continue
                        task = asyncio.create_task(
                            self._tunneld_core.start_tunnel_task(remotepairing_ip, remotepairing, queue=queue),
                            name=f'start-tunnel-wifi-{remotepairing_ip}')
                        self._tunneld_core.tunnel_tasks[remotepairing_ip] = TunnelTask(
                            task=task, udid=remotepairing.remote_identifier)
                        created_task = True
            except Exception as e:
                return fastapi.Response(status_code=501,
                                        content=json.dumps({'error': {
                                            'exception': e.__class__.__name__,
                                            'traceback': traceback.format_exc(),
                                        }}))

            if not created_task:
                return fastapi.Response(status_code=501, content=json.dumps({'error': 'task not not created'}))

            tunnel: Optional[TunnelResult] = await queue.get()
            if tunnel is not None:
                data = {
                    'interface': tunnel.interface,
                    'port': tunnel.port,
                    'address': tunnel.address
                }
                return generate_http_response(data)
            else:
                return fastapi.Response(status_code=404,
                                        content=json.dumps({'error': 'something went wrong during tunnel creation'}))

    def _run_app(self) -> None:
        uvicorn.run(self._app, host=self.host, port=self.port, loop='asyncio')


async def async_get_tunneld_devices(tunneld_address: Tuple[str, int] = TUNNELD_DEFAULT_ADDRESS) \
        -> List[RemoteServiceDiscoveryService]:
    tunnels = _list_tunnels(tunneld_address)
    return await _create_rsds_from_tunnels(tunnels)


def get_tunneld_devices(tunneld_address: Tuple[str, int] = TUNNELD_DEFAULT_ADDRESS) \
        -> List[RemoteServiceDiscoveryService]:
    return get_asyncio_loop().run_until_complete(async_get_tunneld_devices(tunneld_address))


async def async_get_tunneld_device_by_udid(udid: str, tunneld_address: Tuple[str, int] = TUNNELD_DEFAULT_ADDRESS) \
        -> Optional[RemoteServiceDiscoveryService]:
    tunnels = _list_tunnels(tunneld_address)
    if udid not in tunnels:
        return None
    rsds = await _create_rsds_from_tunnels({udid: tunnels[udid]})
    return rsds[0]


def get_tunneld_device_by_udid(udid: str, tunneld_address: Tuple[str, int] = TUNNELD_DEFAULT_ADDRESS) \
        -> Optional[RemoteServiceDiscoveryService]:
    return get_asyncio_loop().run_until_complete(async_get_tunneld_device_by_udid(udid, tunneld_address))


def _list_tunnels(tunneld_address: Tuple[str, int] = TUNNELD_DEFAULT_ADDRESS) -> Mapping[str, List[Mapping]]:
    try:
        # Get the list of tunnels from the specified address
        resp = requests.get(f'http://{tunneld_address[0]}:{tunneld_address[1]}')
        tunnels = resp.json()
    except requests.exceptions.ConnectionError:
        raise TunneldConnectionError()
    return tunnels


async def _create_rsds_from_tunnels(tunnels: Mapping[str, List[Mapping]]) -> List[RemoteServiceDiscoveryService]:
    rsds = []
    for udid, details in tunnels.items():
        for tunnel_details in details:
            rsd = RemoteServiceDiscoveryService((tunnel_details['tunnel-address'], tunnel_details['tunnel-port']),
                                                name=tunnel_details['interface'])
            try:
                await rsd.connect()
                rsds.append(rsd)
            except (TimeoutError, ConnectionError):
                continue
    return rsds



import asyncio
import datetime
import json
import logging
import os
import sys
import uuid
from functools import wraps
from typing import Any, Callable, List, Mapping, Optional, Tuple

import click
import coloredlogs
import hexdump
import inquirer3
from click import Option, UsageError
from inquirer3.themes import GreenPassion
from pygments import formatters, highlight, lexers

from pymobiledevice3.exceptions import AccessDeniedError, DeviceNotFoundError, NoDeviceConnectedError
from pymobiledevice3.lockdown import LockdownClient, create_using_usbmux
from pymobiledevice3.osu.os_utils import get_os_utils
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.tunneld import TUNNELD_DEFAULT_ADDRESS, async_get_tunneld_devices
from pymobiledevice3.usbmux import select_devices_by_connection_type

USBMUX_OPTION_HELP = 'usbmuxd listener address (in the form of either /path/to/unix/socket OR HOST:PORT'
COLORED_OUTPUT = True
UDID_ENV_VAR = 'PYMOBILEDEVICE3_UDID'
OSUTILS = get_os_utils()


class RSDOption(Option):
    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop('mutually_exclusive', []))
        help = kwargs.get('help', '')
        if self.mutually_exclusive:
            ex_str = ', '.join(self.mutually_exclusive)
            kwargs['help'] = help + (
                    '\nNOTE: This argument is mutually exclusive with '
                    ' arguments: [' + ex_str + '].'
            )
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if (isinstance(ctx.command, RSDCommand) and not (isinstance(ctx.command, Command)) and
                ('rsd_service_provider_using_tunneld' not in opts) and ('rsd_service_provider_manually' not in opts)):
            # defaulting to `--tunnel ''` if no remote option was specified
            opts['rsd_service_provider_using_tunneld'] = ''
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise UsageError(
                'Illegal usage: `{}` is mutually exclusive with '
                'arguments `{}`.'.format(
                    self.name,
                    ', '.join(self.mutually_exclusive)
                )
            )

        return super().handle_parse_result(
            ctx,
            opts,
            args
        )


def default_json_encoder(obj):
    if isinstance(obj, bytes):
        return f'<{obj.hex()}>'
    if isinstance(obj, datetime.datetime):
        return str(obj)
    if isinstance(obj, uuid.UUID):
        return str(obj)
    raise TypeError()


def print_json(buf, colored: Optional[bool] = None, default=default_json_encoder):
    if colored is None:
        colored = user_requested_colored_output()
    formatted_json = json.dumps(buf, sort_keys=True, indent=4, default=default)
    if colored and os.isatty(sys.stdout.fileno()):
        colorful_json = highlight(formatted_json, lexers.JsonLexer(),
                                  formatters.TerminalTrueColorFormatter(style='stata-dark'))
        print(colorful_json)
        return colorful_json
    else:
        print(formatted_json)
        return formatted_json


def print_hex(data, colored=True):
    hex_dump = hexdump.hexdump(data, result='return')
    if colored:
        print(highlight(hex_dump, lexers.HexdumpLexer(), formatters.TerminalTrueColorFormatter(style='native')))
    else:
        print(hex_dump, end='\n\n')


def set_verbosity(ctx, param, value):
    coloredlogs.set_level(logging.INFO - (value * 10))


def set_color_flag(ctx, param, value) -> None:
    global COLORED_OUTPUT
    COLORED_OUTPUT = value


def user_requested_colored_output() -> bool:
    return COLORED_OUTPUT and os.isatty(sys.stdout.fileno())


def get_last_used_terminal_formatting(buf: str) -> str:
    return '\x1b' + buf.rsplit('\x1b', 1)[1].split('m')[0] + 'm'


def sudo_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not OSUTILS.is_admin:
            raise AccessDeniedError()
        else:
            func(*args, **kwargs)

    return wrapper


def prompt_selection(choices: List[Any], message: str, idx: bool = False) -> Any:
    question = [inquirer3.List('selection', message=message, choices=choices, carousel=True)]
    try:
        result = inquirer3.prompt(question, theme=GreenPassion(), raise_keyboard_interrupt=True)
    except KeyboardInterrupt:
        raise click.ClickException('No selection was made')
    return result['selection'] if not idx else choices.index(result['selection'])


def prompt_device_list(device_list: List):
    return prompt_selection(device_list, 'Choose device')


def choose_service_provider(callback: Callable):
    def wrap_callback_calling(**kwargs: Mapping):
        service_provider = None
        lockdown_service_provider = kwargs.pop('lockdown_service_provider', None)
        rsd_service_provider_manually = kwargs.pop('rsd_service_provider_manually', None)
        rsd_service_provider_using_tunneld = kwargs.pop('rsd_service_provider_using_tunneld', None)
        if lockdown_service_provider is not None:
            service_provider = lockdown_service_provider
        if rsd_service_provider_manually is not None:
            service_provider = rsd_service_provider_manually
        if rsd_service_provider_using_tunneld is not None:
            service_provider = rsd_service_provider_using_tunneld
        callback(service_provider=service_provider, **kwargs)

    return wrap_callback_calling


class BaseCommand(click.Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params[:0] = [
            click.Option(('verbosity', '-v', '--verbose'), count=True, callback=set_verbosity, expose_value=False),
            click.Option(('color', '--color/--no-color'), default=True, callback=set_color_flag, is_flag=True,
                         expose_value=False, help='colorize output'),
        ]


class BaseServiceProviderCommand(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params[:0] = [
            click.Option(('verbosity', '-v', '--verbose'), count=True, callback=set_verbosity, expose_value=False),
        ]
        self.service_provider = None
        self.callback = choose_service_provider(self.callback)


class LockdownCommand(BaseServiceProviderCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.usbmux_address = None
        self.params[:0] = [
            click.Option(('usbmux', '--usbmux'), callback=self.usbmux, expose_value=False,
                         help=USBMUX_OPTION_HELP),
            click.Option(('lockdown_service_provider', '--udid'), envvar=UDID_ENV_VAR, callback=self.udid,
                         help=f'Device unique identifier. You may pass {UDID_ENV_VAR} environment variable to pass this'
                              f' option as well'),
        ]

    def usbmux(self, ctx, param: str, value: Optional[str] = None) -> None:
        if value is None:
            return
        self.usbmux_address = value

    def udid(self, ctx, param: str, value: str) -> Optional[LockdownClient]:
        if '_PYMOBILEDEVICE3_COMPLETE' in os.environ:
            # prevent lockdown connection establishment when in autocomplete mode
            return

        if self.service_provider is not None:
            return self.service_provider

        if value is not None:
            return create_using_usbmux(serial=value)

        devices = select_devices_by_connection_type(connection_type='USB', usbmux_address=self.usbmux_address)
        if len(devices) <= 1:
            return create_using_usbmux(usbmux_address=self.usbmux_address)

        return prompt_device_list(
            [create_using_usbmux(serial=device.serial, usbmux_address=self.usbmux_address) for device in devices])


class RSDCommand(BaseServiceProviderCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params[:0] = [
            RSDOption(('rsd_service_provider_manually', '--rsd'), type=(str, int), callback=self.rsd,
                      mutually_exclusive=['rsd_service_provider_using_tunneld'],
                      help='\b\n'
                           'RSD hostname and port number (as provided by a `start-tunnel` subcommand).'),
            RSDOption(('rsd_service_provider_using_tunneld', '--tunnel'), callback=self.tunneld,
                      mutually_exclusive=['rsd_service_provider_manually'],
                      help='\b\n'
                           'Either an empty string to force tunneld device selection, or a UDID of a tunneld '
                           'discovered device.\n'
                           'The string may be suffixed with :PORT in case tunneld is not serving at the default port.')
        ]

    def rsd(self, ctx, param: str, value: Optional[Tuple[str, int]]) -> Optional[RemoteServiceDiscoveryService]:
        if value is not None:
            rsd = RemoteServiceDiscoveryService(value)
            asyncio.run(rsd.connect(), debug=True)
            self.service_provider = rsd
            return self.service_provider

    async def _tunneld(self, udid: Optional[str] = None) -> Optional[RemoteServiceDiscoveryService]:
        if udid is None:
            return

        port = TUNNELD_DEFAULT_ADDRESS[1]
        if ':' in udid:
            udid, port = udid.split(':')

        rsds = await async_get_tunneld_devices((TUNNELD_DEFAULT_ADDRESS[0], port))
        if len(rsds) == 0:
            raise NoDeviceConnectedError()

        if udid != '':
            try:
                # Connect to the specified device
                self.service_provider = [rsd for rsd in rsds if rsd.udid == udid][0]
            except IndexError:
                raise DeviceNotFoundError(udid)
        else:
            if len(rsds) == 1:
                self.service_provider = rsds[0]
            else:
                self.service_provider = prompt_device_list(rsds)

        for rsd in rsds:
            if rsd == self.service_provider:
                continue
            await rsd.close()

        return self.service_provider

    def tunneld(self, ctx, param: str, udid: Optional[str] = None) -> Optional[RemoteServiceDiscoveryService]:
        return asyncio.run(self._tunneld(udid), debug=True)


class Command(RSDCommand, LockdownCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CommandWithoutAutopair(Command):
    @staticmethod
    def udid(ctx, param, value):
        if '_PYMOBILEDEVICE3_COMPLETE' in os.environ:
            # prevent lockdown connection establishment when in autocomplete mode
            return
        return create_using_usbmux(serial=value, autopair=False)


class BasedIntParamType(click.ParamType):
    name = 'based int'

    def convert(self, value, param, ctx):
        try:
            return int(value, 0)
        except ValueError:
            self.fail(f'{value!r} is not a valid int.', param, ctx)


BASED_INT = BasedIntParamType()



import logging

import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.amfi import AmfiService

logger = logging.getLogger(__name__)


@click.group()
def cli() -> None:
    pass


@cli.group()
def amfi() -> None:
    """ Enable developer-mode or query its state """
    pass


@amfi.command(cls=Command)
def reveal_developer_mode(service_provider: LockdownClient):
    """ reveal developer mode option in device's UI """
    AmfiService(service_provider).reveal_developer_mode_option_in_ui()


@amfi.command(cls=Command)
def enable_developer_mode(service_provider: LockdownClient):
    """ enable developer mode """
    AmfiService(service_provider).enable_developer_mode()


@amfi.command(cls=Command)
def developer_mode_status(service_provider: LockdownClient):
    """ query developer mode status """
    print_json(service_provider.developer_mode_status)
